"""Script to run Fluent test in Docker container."""

import argparse
import concurrent.futures
import logging
import os
from pathlib import Path
from shutil import copytree
from tempfile import TemporaryDirectory
from time import sleep

import yaml

import ansys.fluent.core as pyfluent
from ansys.fluent.core import FluentVersion
import docker


class FluentRuntimeError(RuntimeError):
    """Exception raised when stderr is detected in Fluent output."""

    pass


def _run_single_test(
    src_test_dir: Path, journal_file: Path, launcher_args: str
) -> None:
    """Run a single Fluent test.

    Parameters
    ----------
    src_test_dir : Path
        Path to the Fluent test directory in the host.

    journal_file : Path
        Absolute path to the journal file.

    launcher_args : str, optional
        Additional arguments for the Fluent launcher.

    Raises
    ------
    FluentRuntimeError
        Raised when stderr is detected in Fluent output.
    """
    logging.debug(f"journal_file: {journal_file}")
    src_pyfluent_dir = str(Path(pyfluent.__file__).parent)
    verion_for_file_name = FluentVersion.current_dev().number
    dst_pyfluent_dir = f"/ansys_inc/v{verion_for_file_name}/commonfiles/CPython/3_10/linx64/Release/python/lib/python3.10/site-packages/ansys/fluent/core"
    src_gen_dir = (
        Path(pyfluent.__file__).parent / "ansys" / "fluent" / "core" / "generated"
    )
    dst_gen_dir = f"/ansys_inc/v{verion_for_file_name}/fluent/fluent{FluentVersion.current_dev()!r}/cortex/pylib/flapi/generated"
    dst_test_dir = "/testing"
    working_dir = Path(dst_test_dir)
    parent = journal_file.parent
    parents = []
    while parent != src_test_dir:
        parents.append(parent.name)
        parent = parent.parent
    parents.reverse()
    for parent in parents:
        working_dir /= parent
    working_dir = str(working_dir)
    src_test_dir = str(src_test_dir)
    logging.debug(f"src_pyfluent_dir: {src_pyfluent_dir}")
    logging.debug(f"dst_pyfluent_dir: {dst_pyfluent_dir}")
    logging.debug(f"src_test_dir: {src_test_dir}")
    logging.debug(f"dst_test_dir: {dst_test_dir}")
    logging.debug(f"working_dir: {working_dir}")

    docker_client = docker.from_env()
    version_for_image_tag = FluentVersion.current_dev().docker_image_tag
    image_name = f"ghcr.io/ansys/pyfluent:{version_for_image_tag}"
    container = docker_client.containers.run(
        image=image_name,
        volumes=[
            f"{src_pyfluent_dir}:{dst_pyfluent_dir}",
            f"{src_gen_dir}:{dst_gen_dir}",  # Try removing this after pyfluent is updated in commonfiles
            f"{src_test_dir}:{dst_test_dir}",
        ],
        working_dir=working_dir,
        environment={"ANSYSLMD_LICENSE_FILE": os.environ["ANSYSLMD_LICENSE_FILE"]},
        command=f"{launcher_args} -gu -py -i {journal_file.name}",
        detach=True,
        stdout=True,
        stderr=True,
        auto_remove=True,
    )
    try:
        while True:
            container.reload()
            if container.status == "exited":
                break
            stderr = container.logs(stdout=False, stderr=True)
            if stderr:
                stderr = stderr.decode()
                for line in stderr.splitlines():
                    if line.strip().startswith("Error:"):
                        if "Expected exception" in line:  # for check_assert.py
                            container.stop()
                        else:
                            raise FluentRuntimeError(stderr)
            sleep(1)
        print(container.logs(stderr=True).decode())
        container.remove()
    except docker.errors.DockerException:
        pass


MAX_TEST_PATH_LENGTH = 100


def _run_single_test_with_status_print(
    src_test_dir: Path, journal_file: Path, launcher_args: str, test_file_relpath: str
) -> bool:
    try:
        _run_single_test(src_test_dir, journal_file, launcher_args)
        print(
            f"{test_file_relpath}{(MAX_TEST_PATH_LENGTH + 10 - len(test_file_relpath)) * '·'}PASSED"
        )
    except FluentRuntimeError as e:
        print(
            f"{test_file_relpath}{(MAX_TEST_PATH_LENGTH + 10 - len(test_file_relpath)) * '·'}FAILED"
        )
        print(e)
        return True


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run Fluent test.")
    parser.add_argument(
        "test_dir",
        help="Path to the Fluent test directory relative to the PyFluent repository root.",
    )
    args = parser.parse_args()
    test_dir = Path.cwd() / args.test_dir
    with TemporaryDirectory(ignore_cleanup_errors=True) as src_test_dir:
        copytree(test_dir, src_test_dir, dirs_exist_ok=True)
        statuses = []
        arguments = []
        src_test_dir = Path(src_test_dir)
        for test_file in (src_test_dir / "fluent").rglob("*.py"):
            config_file = test_file.with_suffix(".yaml")
            launcher_args = ""
            if config_file.exists():
                configs = yaml.safe_load(config_file.read_text())
                launcher_args = configs.get("launcher_args", "")
            test_file_relpath = str(test_file.relative_to(src_test_dir))
            arguments.append(
                (src_test_dir, test_file, launcher_args, test_file_relpath)
            )
        max_workers = int(os.getenv("MAX_WORKERS_FLUENT_TESTS", 4))
        if max_workers > 1:
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=max_workers
            ) as executor:
                futures = [
                    executor.submit(_run_single_test_with_status_print, *args)
                    for args in arguments
                ]
                for future in concurrent.futures.as_completed(futures):
                    statuses.append(future.result())
        else:
            for args in arguments:
                statuses.append(_run_single_test_with_status_print(*args))
        if any(statuses):
            exit(1)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Functions to download sample datasets from the Ansys example data repository."""

import logging
import os
from pathlib import Path
import re
import shutil
import zipfile

import ansys.fluent.core as pyfluent
from ansys.fluent.core.utils.networking import check_url_exists, get_url_content

logger = logging.getLogger("pyfluent.networking")


class RemoteFileNotFoundError(FileNotFoundError):
    """Raised on an attempt to download a non-existent remote file."""

    def __init__(self, url):
        """Initializes RemoteFileNotFoundError."""
        super().__init__(f"{url} does not exist.")


def delete_downloads():
    """Delete all downloaded examples from the default examples folder to free space or
    update the files.

    Notes
    -----
    The default examples path is given by ``pyfluent.EXAMPLES_PATH``.
    """
    shutil.rmtree(pyfluent.EXAMPLES_PATH)
    os.makedirs(pyfluent.EXAMPLES_PATH)


def _decompress(file_name: str) -> None:
    """Decompress zipped file."""
    zip_ref = zipfile.ZipFile(file_name, "r")
    zip_ref.extractall(pyfluent.EXAMPLES_PATH)
    return zip_ref.close()


def _get_file_url(file_name: str, directory: str | None = None) -> str:
    """Get file URL."""
    if directory:
        return (
            "https://github.com/ansys/example-data/raw/master/"
            f"{directory}/{file_name}"
        )
    return f"https://github.com/ansys/example-data/raw/master/{file_name}"


def _retrieve_file(
    url: str,
    file_name: str,
    save_path: str | None = None,
    return_without_path: bool | None = False,
) -> str:
    """Download specified file from specified URL."""
    file_name = os.path.basename(file_name)
    if save_path is None:
        save_path = os.getenv(
            "PYFLUENT_CONTAINER_MOUNT_SOURCE",
            pyfluent.CONTAINER_MOUNT_SOURCE or os.getcwd(),
        )
    else:
        save_path = os.path.abspath(save_path)
    local_path = os.path.join(save_path, file_name)
    local_path_no_zip = re.sub(".zip$", "", local_path)
    file_name_no_zip = re.sub(".zip$", "", file_name)
    # First check if file has already been downloaded
    logger.info(f"Checking if {local_path_no_zip} already exists...")
    if os.path.isfile(local_path_no_zip) or os.path.isdir(local_path_no_zip):
        logger.info("File already exists.")
        if return_without_path:
            return file_name_no_zip
        else:
            return local_path_no_zip

    logger.info("File does not exist. Downloading specified file...")

    # Check if save path exists
    if not os.path.exists(save_path):
        os.makedirs(save_path)

    # Download file
    logger.info(f'Downloading URL: "{url}"')
    content = get_url_content(url)
    with open(local_path, "wb") as f:
        f.write(content)

    if local_path.endswith(".zip"):
        _decompress(local_path)
        local_path = local_path_no_zip
        file_name = file_name_no_zip
    logger.info("Download successful.")
    if return_without_path:
        return file_name
    else:
        return local_path


def download_file(
    file_name: str,
    directory: str | None = None,
    save_path: str | None = None,
    return_without_path: bool | None = None,
) -> str:
    """Download specified example file from the Ansys example data repository.

    Parameters
    ----------
    file_name : str
        File to download.
    directory : str, optional
        Ansys example data repository directory where specified file is located. If not specified, looks for the file
        in the root directory of the repository.
    save_path : str, optional
        Path to download the specified file to.
    return_without_path : bool, optional
        When unspecified, defaults to False, unless the PYFLUENT_LAUNCH_CONTAINER=1 environment variable is specified,
        in which case defaults to True.
        Relevant when using Fluent Docker container images, as the full path for the imported file from
        the host side is not necessarily going to be the same as the one for Fluent inside the container.
        Assuming the Fluent inside the container has its working directory set to the path that was mounted from
        the host, and that the example files are being made available by the host through this same path,
        only the file name is required for Fluent to find and open the file.

    Raises
    ------
    RemoteFileNotFoundError
        If remote file does not exist.

    Returns
    -------
    str
        file path of the downloaded or already existing file, or only the file name if ``return_without_path=True``.

    Examples
    --------
    >>> from ansys.fluent.core import examples
    >>> file_path = examples.download_file("bracket.iges", "geometry")
    >>> file_path
    '/home/user/.local/share/ansys_fluent_core/examples/bracket.iges'
    >>> file_name = examples.download_file("bracket.iges", "geometry", return_without_path=True)
    >>> file_name
    'bracket.iges'
    >>> file_path = examples.download_file("bracket.iges", "geometry", save_path='.')
    '/home/<current_folder_path>/bracket.iges'
    >>> file_name = examples.download_file("bracket.iges", "geometry", save_path='.', return_without_path=True)
    >>> file_name
    'bracket.iges'
    >>> file_path = examples.download_file("bracket.iges", "geometry", save_path='<user_specified_path>')
    '/home/<user_specified_path>/bracket.iges'
    >>> file_name = examples.download_file("bracket.iges", "geometry", save_path='<user_specified_path>',
    ...                                   return_without_path=True)
    >>> file_name
    'bracket.iges'
    """
    if return_without_path is None:
        if os.getenv("PYFLUENT_LAUNCH_CONTAINER") == "1":
            if pyfluent.USE_FILE_TRANSFER_SERVICE:
                return_without_path = False
            else:
                return_without_path = True

    url = _get_file_url(file_name, directory)
    if not check_url_exists(url):
        raise RemoteFileNotFoundError(url)
    return _retrieve_file(url, file_name, save_path, return_without_path)


def path(file_name: str):
    """Return path of given file name.

    Parameters
    ----------
    file_name : str
        Name of the file.

    Raises
    ------
    FileNotFoundError
        If file does not exist.

    Returns
    -------
    file_path: str
        File path.
    """
    if os.path.isabs(file_name):
        return file_name
    file_path = Path(pyfluent.EXAMPLES_PATH) / file_name
    if file_path.is_file():
        return str(file_path)
    else:
        raise FileNotFoundError(f"{file_name} does not exist.")



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from os.path import dirname, join
from pathlib import Path
import shutil

import defusedxml.ElementTree as ET
import pytest

from ansys.fluent.core import examples
from ansys.fluent.core.filereader import lispy
from ansys.fluent.core.filereader.case_file import (
    InputParameter,
    InputParameterOld,
    MeshType,
    _get_processed_string,
)
from ansys.fluent.core.filereader.case_file import CaseFile
from ansys.fluent.core.filereader.case_file import CaseFile as CaseReader
from ansys.fluent.core.filereader.pre_processor import remove_unsupported_xml_chars


def call_casereader(
    case_file_name: str | None = None,
    project_file_name: str | None = None,
    expected: dict | None = None,
):
    reader = CaseReader(
        case_file_name=case_file_name, project_file_name=project_file_name
    )
    if expected is not None:
        assert reader.precision() == expected["precision"]
        assert reader.num_dimensions() == expected["num_dimensions"]
        assert reader.iter_count() == expected["iter_count"]
        assert {
            p.name: (p.numeric_value, p.units) for p in reader.input_parameters()
        } == expected["input_parameters"]
        assert {p.name: p.units for p in reader.output_parameters()} == expected[
            "output_parameters"
        ]


def call_casereader_static_mixer(
    case_file_name: str | None = None, project_file_name: str | None = None
):
    call_casereader(
        case_file_name=case_file_name,
        project_file_name=project_file_name,
        expected=dict(
            precision=2,
            num_dimensions=3,
            iter_count=100,
            input_parameters=dict(
                inlet1_temp=(300, "K"),
                inlet1_vel=(1, "m/s"),
                inlet2_temp=(350, "K"),
                inlet2_vel=(1, "m/s"),
            ),
            output_parameters={
                "outlet-temp-avg-op": "K",
                "outlet-vel-avg-op": "m s^-1",
            },
        ),
    )


def static_mixer_file():
    return examples.download_file(
        "Static_Mixer_Parameters.cas.h5",
        "pyfluent/static_mixer",
        return_without_path=False,
    )


def test_casereader_static_mixer_h5():
    call_casereader_static_mixer(case_file_name=static_mixer_file())


def test_casereader_static_mixer_binary_cas():
    call_casereader_static_mixer(
        case_file_name=examples.download_file(
            "Static_Mixer_Parameters_legacy_binary.cas",
            "pyfluent/static_mixer",
            return_without_path=False,
        )
    )


def test_casereader_static_mixer_binary_gz():
    call_casereader_static_mixer(
        case_file_name=examples.download_file(
            "Static_Mixer_Parameters_legacy_binary.cas.gz",
            "pyfluent/static_mixer",
            return_without_path=False,
        )
    )


def test_casereader_static_mixer_text_cas():
    call_casereader_static_mixer(
        case_file_name=examples.download_file(
            "Static_Mixer_Parameters_legacy_text.cas",
            "pyfluent/static_mixer",
            return_without_path=False,
        )
    )


def test_casereader_static_mixer_text_gz():
    call_casereader_static_mixer(
        case_file_name=examples.download_file(
            "Static_Mixer_Parameters_legacy_text.cas.gz",
            "pyfluent/static_mixer",
            return_without_path=False,
        )
    )


def create_dir_structure_locally(copy_1: bool = False, copy_2: bool = False):
    # Copying from and then creating the entire directory structure locally
    case_file_dir = (
        "Static_Mixer_Parameter_project_file/"
        "Static_Mixer_Parameters.cffdb/Static_Mixer_Parameters-Solve"
    )
    case_file_name = examples.download_file(
        "Static_Mixer_Parameters.cas.h5",
        "pyfluent/static_mixer/" + case_file_dir,
        return_without_path=False,
    )
    prj_dir = join(dirname(case_file_name), case_file_dir)
    Path(prj_dir).mkdir(parents=True, exist_ok=True)
    if copy_1:
        shutil.copy2(case_file_name, prj_dir)
    if copy_2:
        case_file_name_2 = examples.download_file(
            "Static_Mixer_Parameters_legacy_binary.cas.gz",
            "pyfluent/static_mixer",
            return_without_path=False,
        )
        shutil.copy2(case_file_name_2, prj_dir)
    prj_file_dir = "Static_Mixer_Parameter_project_file"
    prj_file = r"Static_Mixer_Parameters.flprj"
    prj_file_name = examples.download_file(
        prj_file, "pyfluent/static_mixer/" + prj_file_dir, return_without_path=False
    )
    prj_file_dir = join(dirname(prj_file_name), prj_file_dir)
    shutil.copy2(prj_file_name, prj_file_dir)

    return join(prj_file_dir, prj_file)


def test_processed_string():
    assert (
        _get_processed_string(b"Hello! World (37 ( Get this part of the string ))")
        == "(37 ( Get this part of the string ))"
    )


def test_casereader_no_file():
    with pytest.raises(FileNotFoundError):
        call_casereader("no_file.cas.h5")


def test_casereader_with_both_project_and_case_file():
    with pytest.raises(RuntimeError):
        call_casereader(
            case_file_name="case_file.cas.h5", project_file_name="project.flprj"
        )


def test_casereader_for_project_directory_invalid_project_file():
    with pytest.raises(FileNotFoundError):
        call_casereader(project_file_name="project.flprx")


def test_case_reader_with_bad_data_to_be_skipped_and_input_parameters_labeled_differently():
    call_casereader(
        case_file_name=examples.download_file(
            "mixer-ran_2019r3.cas.gz", "pyfluent/optislang", return_without_path=False
        ),
        expected=dict(
            precision=1,
            num_dimensions=3,
            iter_count=25,
            input_parameters=dict(
                swirl_max_hot=(0.1, "m s^-1"),
                vel_hot=(0.1, "m s^-1"),
                vel_cold=(0.1, "m s^-1"),
                swirl_max_cold=(0.1, "m s^-1"),
            ),
            output_parameters={
                "p2-op": "kg m^-1 s^-2",
                "t-dev-op": "K",
                "p1-op": "kg m^-1 s^-2",
                "ave_temp_out": "K",
            },
        ),
    )


def test_case_reader_get_rp_and_config_vars():
    reader = CaseReader(case_file_name=static_mixer_file())
    rp_vars = reader.rp_vars()
    assert rp_vars
    assert hasattr(rp_vars, "__getitem__")
    config_vars = reader.config_vars()
    assert config_vars
    assert hasattr(config_vars, "__getitem__")
    assert config_vars["rp-3d?"] is True
    assert reader.config_var("rp-3d?") is True
    assert reader.config_var.rp_3d__q() is True
    assert reader.rp_var.smooth_mesh.niter() == 4
    assert reader.rp_var.pressure.output_dpdt__q() is True
    assert len(reader.rp_var.context.map_r17__plus()) == 53
    assert reader.rp_var.defaults.pre_r19__dot0_early__q() is False

    with pytest.raises(RuntimeError) as msg:
        reader.rp_var.defaults.pre_r19__dot0_early()

    with pytest.raises(ValueError) as msg:
        reader.config_var("rp-3d")
    assert (
        msg.value.args[0] == "'config-vars' has no attribute 'rp-3d'.\n"
        "The most similar names are: rp-3d?, rp-des?."
    )


def test_case_reader_input_parameter():
    number = InputParameter(raw_data=(("name", "n"), ("definition", "12.4")))

    assert number.name == "n"
    assert number.units == ""
    assert number.numeric_value == 12.4
    assert number.value == "12.4"

    length = InputParameter(raw_data=(("name", "x"), ("definition", "12.4 [m]")))

    assert length.name == "x"
    assert length.units == "m"
    assert length.numeric_value == 12.4
    assert length.value == "12.4 [m]"

    momentum = InputParameter(
        raw_data=(("name", "p"), ("definition", "12.4 [kg m s^-1]"))
    )

    assert momentum.name == "p"
    assert momentum.units == "kg m s^-1"
    assert momentum.numeric_value == 12.4
    assert momentum.value == "12.4 [kg m s^-1]"

    velocity = InputParameter(raw_data=(("name", "v"), ("definition", "2[m/s]")))

    assert velocity.name == "v"
    assert velocity.units == "m/s"
    assert velocity.numeric_value == 2
    assert velocity.value == "2 [m/s]"

    vel_data = [
        "real-1",
        [
            ("type", "real"),
            ["name", ("value", '"inlet_velocity"'), ("type", "string-class")],
            [
                "parameter-value",
                ("type", "real-class"),
                ("value", 20.0),
                ("min", False),
                ("max", False),
                ("units-quantity", "velocity"),
            ],
        ],
    ]
    velocity = InputParameterOld(raw_data=vel_data)

    assert velocity.name == "inlet_velocity"
    assert velocity.value == [
        "parameter-value",
        ("type", "real-class"),
        ("value", 20.0),
        ("min", False),
        ("max", False),
        ("units-quantity", "velocity"),
    ]
    assert velocity.units == "velocity"
    assert velocity.numeric_value == 20.0


def test_lispy_for_multiline_string():
    assert lispy.parse('(define x "abc\ndef")') == ["define", "x", '"abc\ndef"']


def test_lispy_for_quotes():
    assert lispy.parse(
        '(define x "\n(format \\"\n-------------------------\nRunning Original Settings\n------------------------\n\\")")'
    ) == [
        "define",
        "x",
        '"\n(format \\"\n-------------------------\nRunning Original Settings\n------------------------\n\\")"',
    ]


def test_mesh_reader():
    mesh_file_2d = examples.download_file(
        "sample_2d_mesh.msh.h5",
        "pyfluent/surface_mesh",
        return_without_path=False,
    )
    mesh_file_3d = examples.download_file(
        "mixing_elbow.msh.h5",
        "pyfluent/mixing_elbow",
        return_without_path=False,
    )
    case_file = examples.download_file(
        "mixing_elbow.cas.h5",
        "pyfluent/mixing_elbow",
        return_without_path=False,
    )
    mesh_reader_2d = CaseReader(case_file_name=mesh_file_2d)
    mesh_reader_3d = CaseReader(case_file_name=mesh_file_3d)
    case_reader = CaseReader(case_file_name=case_file)

    assert mesh_reader_2d.get_mesh().get_mesh_type() == MeshType.SURFACE
    assert mesh_reader_3d.get_mesh().get_mesh_type() == MeshType.VOLUME
    assert case_reader.get_mesh().get_mesh_type() == MeshType.VOLUME

    assert mesh_reader_2d.precision() is None
    assert mesh_reader_3d.precision() is None
    assert case_reader.precision() == 2


def test_preprocessor():
    content = """
    <Project type="object" class="PFolder">
    <Metadata type="object" class="ansys::Project::MetadataHoof">
        <ProjectStoragePolicy::FolderEnabled class="string" value="true"/>
    </Metadata>
    </Project>
    """
    expected = """
    <Project type="object" class="PFolder">
    <Metadata type="object" class="ansys::Project::MetadataHoof">
        <ProjectStoragePolicy__FolderEnabled class="string" value="true"/>
    </Metadata>
    </Project>
    """
    with pytest.raises(ET.ParseError):
        ET.fromstring(content)
    pre_processed = remove_unsupported_xml_chars(content)
    assert pre_processed == expected
    assert ET.fromstring(pre_processed)


def test_read_flprj_3891():
    # Read the .flprj file from https://github.com/ansys/pyfluent/issues/3891
    data_zip = examples.download_file(
        "data.zip",
        "pyfluent/flprj_3891",
        return_without_path=False,
    )
    prj_file_name = next(Path(data_zip).with_suffix("").glob("*.flprj"))
    assert (
        CaseFile(project_file_name=prj_file_name).get_mesh().get_mesh_type()
        == MeshType.VOLUME
    )



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module to manage datamodel cache."""

from collections import abc, defaultdict
from contextlib import contextmanager
import copy
from enum import Enum
from threading import RLock
from typing import Any, Dict, List, Optional

from ansys.api.fluent.v0.variant_pb2 import Variant
from ansys.fluent.core.utils.fluent_version import FluentVersion

StateType = (
    bool
    | int
    | float
    | str
    | List[bool]
    | List[int]
    | List[float]
    | List[str]
    | List["StateType"]
    | Dict[str, "StateType"]
)


class NameKey(Enum):
    """Name key."""

    INTERNAL = "__iname__"
    DISPLAY = "_name_"

    def __invert__(self):
        lst = list(NameKey)
        return lst[~lst.index(self)]


class _CacheImpl:
    def __init__(self, name_key: NameKey):
        self.name_key = name_key

    @staticmethod
    def add_missing_name_keys(k: str, v: dict[str, Any]):
        """Add missing name keys in dict."""
        if ":" in k:
            name_in_key = k.split(":")[1]
            if NameKey.DISPLAY.value in v and v[NameKey.DISPLAY.value] != name_in_key:
                v[NameKey.INTERNAL.value] = name_in_key
            if NameKey.INTERNAL.value in v and v[NameKey.INTERNAL.value] != name_in_key:
                v[NameKey.DISPLAY.value] = name_in_key

    def find(self, d: dict[str, Any], key: str, default: Any) -> tuple[str, Any]:
        """Find in dict."""
        if key in d:
            return key, d[key]
        if ":" in key:
            type_, name = key.split(":")
            for k, v in d.items():
                if (
                    isinstance(v, abc.Mapping)
                    and ":" in k
                    and type_ == k.split(":")[0]
                    and name == v.get((~self.name_key).value, None)
                ):
                    return k, v
            return None, default
        return key, default

    def transform(self, d_in: dict[str, Any], add_missing_name_keys=False):
        """Transform dict."""
        d_out = {}
        for k_in, v_in in d_in.items():
            if isinstance(v_in, abc.Mapping):
                k_out = (
                    f'{k_in.split(":")[0]}:{v_in[(~self.name_key).value]}'
                    if ":" in k_in
                    else k_in
                )
                v_out = self.transform(v_in, add_missing_name_keys)
                if add_missing_name_keys:
                    _CacheImpl.add_missing_name_keys(k_in, v_out)
                d_out[k_out] = v_out
            else:
                d_out[k_in] = v_in
        return d_out

    def update(self, d: dict[str, Any], d1: dict[str, Any]):
        """Update dict."""
        for k1, v1 in d1.items():
            k, v = self.find(d, k1, None)
            if isinstance(v, abc.Mapping) and isinstance(v1, abc.Mapping):
                self.update(v, v1)
            else:
                if isinstance(v1, abc.Mapping):
                    k = (
                        f'{k1.split(":")[0]}:{v1.get(self.name_key.value, k1.split(":")[1])}'
                        if ":" in k1
                        else k1
                    )
                    v1 = _CacheImpl(~self.name_key).transform(v1, True)
                    _CacheImpl.add_missing_name_keys(k1, v1)
                d[k] = v1


def _is_dict_parameter_type(version: FluentVersion, rules: str, rules_path: str):
    """Check if a parameter is a dict type."""
    from ansys.fluent.core import CODEGEN_OUTDIR
    from ansys.fluent.core.services.datamodel_se import (
        PyDictionary,
        PyNamedObjectContainer,
        PyParameter,
    )
    from ansys.fluent.core.utils import load_module

    try:
        module = load_module(
            rules, CODEGEN_OUTDIR / f"datamodel_{version.number}" / f"{rules}.py"
        )
    except FileNotFoundError:  # no codegen or during codegen
        return False
    cls = module.Root
    comps = rules_path.split("/")
    for i, comp in enumerate(comps):
        if hasattr(cls, comp):
            cls = getattr(cls, comp)
            if issubclass(cls, PyParameter) and i < len(comps) - 1:
                return False
            if issubclass(cls, PyNamedObjectContainer):
                cls = getattr(cls, f"_{comp}")
    return issubclass(cls, PyDictionary)


class DataModelCache:
    """Class to manage datamodel cache."""

    use_display_name = False

    def __init__(self):
        """Initialize datamodel cache."""
        self.rules_str_to_cache = defaultdict(dict)
        self.rules_str_to_config = {}
        self._locks = {}

    @contextmanager
    def _with_lock(self, rules: str):
        if rules not in self._locks:
            self._locks[rules] = RLock()
        with self._locks[rules]:
            yield

    class Empty:
        """Class representing unassigned cached state."""

    @staticmethod
    def is_unassigned(state: Any) -> bool:
        """Check whether a cached state is unassigned.

        Parameters
        ----------
        state : Any
            state

        Returns
        -------
        bool
            whether a cached state is unassigned
        """
        return state is DataModelCache.Empty

    def get_config(self, rules: str, name: str) -> Any:
        """Get datamodel cache configuration value.

        Parameters
        ----------
        rules : str
            datamodel rules
        name : str
            configuration name

        Returns
        -------
        Any
            configuration value
        """
        return self.rules_str_to_config.get(rules, {}).get(name, False)

    def set_config(self, rules: str, name: str, value: Any):
        """Set datamodel cache configuration value.

        Parameters
        ----------
        rules : str
            datamodel rules
        name : str
            configuration name
        value : Any
            configuration value
        """
        if rules not in self.rules_str_to_config:
            self.rules_str_to_config[rules] = {}
        self.rules_str_to_config[rules][name] = value

    def _update_cache_from_variant_state(
        self,
        rules: str,
        source: Dict[str, StateType],
        key: str,
        state: Variant,
        updater_fn,
        rules_str: str,
        version,
    ):
        # Helper function to update the source with the state value
        def update_source_with_state(state_field):
            if state.HasField(state_field):
                updater_fn(source, key, getattr(state, state_field))
                return True
            return False

        # Check for basic state types
        for state_type in [
            "bool_state",
            "int64_state",
            "double_state",
            "string_state",
            "bool_vector_state",
            "int64_vector_state",
            "double_vector_state",
            "string_vector_state",
        ]:
            if update_source_with_state(state_type):
                return

        # Handle variant vector state
        if state.HasField("variant_vector_state"):
            updater_fn(source, key, [])
            for item in state.variant_vector_state.item:
                self._update_cache_from_variant_state(
                    rules,
                    source,
                    key,
                    item,
                    lambda d, k, v: d[k].append(v),
                    rules_str + "/" + key.split(":", maxsplit=1)[0],
                    version,
                )
            return

        # Handle variant map state
        if state.HasField("variant_map_state"):
            internal_names_as_keys = (
                self.get_config(rules, "name_key") == NameKey.INTERNAL
            )

            # Determine the appropriate key
            if ":" in key:
                type_, iname = key.split(":", maxsplit=1)
                key = self._determine_key(
                    source, internal_names_as_keys, key, state, type_, iname
                )
            else:
                if key not in source:
                    source[key] = {}

            if version and _is_dict_parameter_type(version, rules, rules_str):
                source[key] = {}

            # Update the source with items from the variant map state
            if state.variant_map_state.item:
                source = source[key]
                for k, v in state.variant_map_state.item.items():
                    self._update_cache_from_variant_state(
                        rules,
                        source,
                        k,
                        v,
                        dict.__setitem__,
                        rules_str + "/" + k.split(":", maxsplit=1)[0],
                        version,
                    )
            else:
                source[key] = {}

        # Default case when no fields are matched
        else:
            updater_fn(source, key, None)

    def _determine_key(
        self,
        source: Dict[str, StateType],
        internal_names_as_keys: bool,
        key: str,
        state: Variant,
        type_: str,
        iname: str,
    ) -> str:
        """Determine the appropriate key based on internal naming conventions."""
        for k1, v1 in source.items():
            if (internal_names_as_keys and k1 == key) or (
                (not internal_names_as_keys)
                and isinstance(v1, dict)
                and v1.get(NameKey.INTERNAL.value) == iname
            ):
                return k1  # Found a matching key

        # If no match found and internal naming is used
        if internal_names_as_keys:
            source[key] = {}
            return key

        # If no match found and external naming is used
        name = state.variant_map_state.item[NameKey.DISPLAY.value].string_state
        new_key = f"{type_}:{name}"
        source[new_key] = {NameKey.INTERNAL.value: iname}

        return new_key

    def update_cache(
        self, rules: str, state: Variant, deleted_paths: List[str], version=None
    ):
        """Update datamodel cache from streamed state.

        Parameters
        ----------
        rules : str
            datamodel rules
        state : Variant
            streamed state
        deleted_paths : List[str]
            list of deleted paths
        version : FluentVersion, optional
            Fluent version
        """
        cache = self.rules_str_to_cache[rules]

        with self._with_lock(rules):
            internal_names_as_keys = (
                self.get_config(rules, "name_key") == NameKey.INTERNAL
            )

            # Process deleted paths
            self._process_deleted_paths(cache, deleted_paths, internal_names_as_keys)

            # Update cache with new state items
            for k, v in state.variant_map_state.item.items():
                self._update_cache_from_variant_state(
                    rules,
                    cache,
                    k,
                    v,
                    dict.__setitem__,
                    k.split(":", maxsplit=1)[0],
                    version,
                )

    def _process_deleted_paths(
        self,
        cache: Dict[str, Any],
        deleted_paths: List[str],
        internal_names_as_keys: bool,
    ):
        """Process and delete paths from the cache based on the deleted paths list."""
        for deleted_path in deleted_paths:
            comps = [x for x in deleted_path.split("/") if x]
            self._delete_from_cache(cache, comps, internal_names_as_keys)

    def _delete_from_cache(
        self, sub_cache: Dict[str, Any], comps: List[str], internal_names_as_keys: bool
    ):
        """Recursively delete components from the cache."""
        for i, comp in enumerate(comps):
            if ":" in comp:
                _, iname = comp.split(":", maxsplit=1)
                key_to_del = self._find_key_to_delete(
                    sub_cache, comp, iname, i == len(comps) - 1, internal_names_as_keys
                )
                if key_to_del:
                    del sub_cache[key_to_del]
                    return  # Exit after deletion
            else:
                if comp in sub_cache:
                    sub_cache = sub_cache[comp]
                else:
                    break

    def _find_key_to_delete(
        self,
        sub_cache: Dict[str, Any],
        comp: str,
        iname: str,
        is_last_component: bool,
        internal_names_as_keys: bool,
    ) -> Optional[str]:
        """Find the key to delete from the sub-cache."""
        for k, v in sub_cache.items():
            if (internal_names_as_keys and k == comp) or (
                (not internal_names_as_keys)
                and isinstance(v, dict)
                and v.get(NameKey.INTERNAL.value) == iname
            ):
                return (
                    k if is_last_component else None
                )  # Return key if it's the last component
        return None  # No key found to delete

    @staticmethod
    def _dm_path_comp(comp):
        return ":".join(comp) if comp[1] else comp[0]

    @staticmethod
    def _dm_path_comp_list(obj):
        return [DataModelCache._dm_path_comp(comp) for comp in obj.path]

    def get_state(
        self, rules: str, obj: object, name_key: NameKey | None = None
    ) -> Any:
        """Retrieve state from datamodel cache.

        Parameters
        ----------
        rules : str
            datamodel rules
        obj : object
            datamodel object, optional
        name_key : NameKey, optional
            if NameKey.INTERNAL, the returned state will contain internal names in keys.
            if NameKey.DISPLAY, the returned state will contain display names in keys.
            Default value is picked from configuration.

        Returns
        -------
        Any
            cached state
        """
        name_key_in_config = self.get_config(rules, "name_key")
        if name_key is None:
            name_key = name_key_in_config
        cache = self.rules_str_to_cache[rules]
        with self._with_lock(rules):
            if not len(cache):
                return DataModelCache.Empty
            comps = DataModelCache._dm_path_comp_list(obj)
            for comp in comps:
                if name_key == name_key_in_config:
                    cache = cache.get(comp, None)
                else:
                    _, cache = _CacheImpl(name_key_in_config).find(cache, comp, None)
                if cache is None:
                    return DataModelCache.Empty

            if not isinstance(cache, abc.Mapping) or name_key == name_key_in_config:
                return copy.deepcopy(cache)
            else:
                if not cache:
                    return DataModelCache.Empty
                return _CacheImpl(name_key_in_config).transform(cache)

    def set_state(self, rules: str, obj: object, value: Any):
        """Set datamodel cache state.

        Parameters
        ----------
        rules : str
            datamodel rules
        obj : object
            datamodel object
        value : Any
            state
        """
        name_key_in_config = self.get_config(rules, "name_key")
        cache = self.rules_str_to_cache[rules]
        with self._with_lock(rules):
            comps = DataModelCache._dm_path_comp_list(obj)
            for i, comp in enumerate(comps):
                key, next_cache = _CacheImpl(name_key_in_config).find(cache, comp, None)
                if i == len(comps) - 1 and not isinstance(value, abc.Mapping):
                    cache[key] = value
                    return
                if isinstance(next_cache, abc.Mapping):
                    cache = next_cache
                else:
                    cache[key] = {}
                    cache = cache[key]
            _CacheImpl(name_key_in_config).update(cache, value)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module for launching Fluent within a Slurm environment.

Examples
--------
Note that the keys ``scheduler_headnode``, ``scheduler_queue`` and ``scheduler_account``
are optional and should be specified in a similar manner to Fluent's scheduler options.

>>> slurm = pyfluent.launch_fluent(
...   scheduler_options={
...     "scheduler": "slurm",
...     "scheduler_headnode": "<headnode>",
...     "scheduler_queue": "<queue>",
...     "scheduler_account": "<account>"
...   },
...   additional_arguments="-t16 -cnf=m1:8,m2:8",
... )
>>> type(slurm)
<class 'ansys.fluent.core.launcher.slurm_launcher.SlurmFuture'>
>>> slurm.pending(), slurm.running(), slurm.done() # before Fluent is launched
(True, False, False)
>>> slurm.pending(), slurm.running(), slurm.done() # after Fluent is launched
(False, True, False)
>>> session = slurm.result()
>>> type(session)
<class 'ansys.fluent.core.session_solver.Solver'>
>>> session.exit()
>>> slurm.pending(), slurm.running(), slurm.done()
(False, False, True)

# Callable slurm launcher

>>> from ansys.fluent.core.launcher.launcher import create_launcher
>>> from ansys.fluent.core.launcher.pyfluent_enums import LaunchMode, FluentMode

>>> slurm_meshing_launcher = create_launcher(LaunchMode.SLURM, mode=FluentMode.MESHING)
>>> slurm_meshing_session = slurm_meshing_launcher()

>>> slurm_solver_launcher = create_launcher(LaunchMode.SLURM)
>>> slurm_solver_session = slurm_solver_launcher()
"""

from concurrent.futures import Future, ThreadPoolExecutor
import inspect
import logging
import os
from pathlib import Path
import shutil
import subprocess
import time
from typing import Any, Callable, Dict

from ansys.fluent.core.exceptions import InvalidArgument
from ansys.fluent.core.launcher.launcher_utils import (
    _await_fluent_launch,
    _build_journal_argument,
    _get_subprocess_kwargs_for_fluent,
)
from ansys.fluent.core.launcher.process_launch_string import _generate_launch_string
from ansys.fluent.core.launcher.pyfluent_enums import (
    Dimension,
    FluentLinuxGraphicsDriver,
    FluentMode,
    FluentWindowsGraphicsDriver,
    Precision,
    UIMode,
    _get_argvals_and_session,
)
from ansys.fluent.core.launcher.server_info import _get_server_info_file_names
from ansys.fluent.core.session_meshing import Meshing
from ansys.fluent.core.session_pure_meshing import PureMeshing
from ansys.fluent.core.session_solver import Solver
from ansys.fluent.core.session_solver_icing import SolverIcing
from ansys.fluent.core.utils.fluent_version import FluentVersion

logger = logging.getLogger("pyfluent.launcher")


def _get_slurm_job_id(proc: subprocess.Popen):
    prefix = "Submitted batch job"
    for line in proc.stdout:
        if line.startswith(prefix.encode()):
            line = line.decode().removeprefix(prefix).strip()
            return int(line)


class _SlurmWrapper:
    """A class wrapping Slurm commands."""

    @staticmethod
    def is_available() -> bool:
        """Check whether Slurm is available.

        Returns
        -------
        bool
            ``True`` if Slurm is available, otherwise ``False``.
        """
        return shutil.which("sinfo") and len(_SlurmWrapper.list_queues()) > 0

    @staticmethod
    def list_queues() -> list[str]:
        """Return list of queues.

        Returns
        -------
        list[str]
        List of queues.
        """
        out = subprocess.check_output(["sinfo", "--format=%P", "--noheader"])
        queues = out.decode().strip().split()
        queues = [q.removesuffix("*") for q in queues]
        return queues

    @staticmethod
    def get_state(job_id: int) -> str:
        """Return state of a job.

        Parameters
        ----------
        job_id : int
            Job id.

        Returns
        -------
        str
            Any of ``""``, ``"RUNNING"``, ``"CANCELLED"`` or ``"COMPLETED"``.
        """
        out = subprocess.check_output(["squeue", "-j", f"{job_id}", "-o", '"%T"', "-h"])
        return out.decode().strip().strip('"')

    @staticmethod
    def cancel(job_id: int) -> None:
        """Cancel a job.

        Parameters
        ----------
        job_id : int
            Job id.
        """
        subprocess.run(["scancel", f"{job_id}"])


class SlurmFuture:
    """Encapsulates asynchronous launch of Fluent within a Slurm environment.

    The interface is similar to Python's
    `future object <https://docs.python.org/3/library/asyncio-future.html#future-object>`_.
    """

    def __init__(self, future: Future, job_id: int):
        """Initialize SlurmFuture."""
        self._future = future
        self._job_id = job_id

    def __enter__(self):
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any):
        self.cancel()

    def _get_state(self) -> str:
        return _SlurmWrapper.get_state(self._job_id)

    def _cancel(self):
        _SlurmWrapper.cancel(self._job_id)

    def cancel(self, timeout: int = 60) -> bool:
        """Attempt to cancel the Fluent launch within timeout seconds.

        Parameters
        ----------
        timeout : int, optional
            timeout in seconds, by default 60

        Returns
        -------
        bool
            ``True`` if the Fluent launch is successfully cancelled, otherwise ``False``.
        """
        if self.done():
            return False
        self._cancel()
        for _ in range(timeout):
            if self._get_state() in ["", "CANCELLED"]:
                return True
            time.sleep(1)
        return False

    def running(self) -> bool:
        """Return ``True`` if Fluent is currently running, otherwise ``False``."""
        return self._get_state() == "RUNNING" and self._future.done()

    def pending(self) -> bool:
        """Return ``True`` if the Fluent launch is currently waiting for Slurm
        allocation or Fluent is being launched, otherwise ``False``."""
        return self._future.running()

    def done(self) -> bool:
        """Return ``True`` if the Fluent launch was successfully cancelled or Fluent was
        finished running, otherwise ``False``."""
        return self._get_state() in ["", "CANCELLED", "COMPLETED"]

    def result(
        self, timeout: int = None
    ) -> Meshing | PureMeshing | Solver | SolverIcing:
        """Return the session instance corresponding to the Fluent launch. If Fluent
        hasn't yet launched, then this method will wait up to timeout seconds. If Fluent
        hasn't launched in timeout seconds, then a TimeoutError will be raised. If
        timeout is not specified or None, there is no limit to the wait time.

        If the Fluent launch raised an exception, this method will raise the same exception.

        Parameters
        ----------
        timeout : int, optional
            timeout in seconds

        Returns
        -------
        :obj:`~typing.Union` [:class:`Meshing<ansys.fluent.core.session_meshing.Meshing>`, \
        :class:`~ansys.fluent.core.session_pure_meshing.PureMeshing`, \
        :class:`~ansys.fluent.core.session_solver.Solver`, \
        :class:`~ansys.fluent.core.session_solver_icing.SolverIcing`]
            The session instance corresponding to the Fluent launch.
        """
        return self._future.result(timeout)

    def exception(self, timeout: int = None) -> Exception:
        """Return the exception raised by the Fluent launch. If Fluent hasn't yet
        launched, then this method will wait up to timeout seconds. If Fluent hasn't
        launched in timeout seconds, then a TimeoutError will be raised. If timeout is
        not specified or None, there is no limit to the wait time.

        If the Fluent launch completed without raising, None is returned.

        Parameters
        ----------
        timeout : int, optional
            timeout in seconds

        Returns
        -------
        Exception
            The exception raised by the Fluent launch.
        """
        return self._future.exception(timeout)

    def add_done_callback(self, fn: Callable):
        """Attaches the callable function. The function will be called, with the session
        as its only argument, when Fluent is launched.

        Parameters
        ----------
        fn : Callable
            Callback function.
        """
        self._future.add_done_callback(fn)


class SlurmLauncher:
    """Instantiates Fluent session within a Slurm environment."""

    def __init__(
        self,
        mode: FluentMode | str | None = None,
        ui_mode: UIMode | str | None = None,
        graphics_driver: (
            FluentWindowsGraphicsDriver | FluentLinuxGraphicsDriver | str | None
        ) = None,
        product_version: FluentVersion | str | float | int | None = None,
        dimension: Dimension | int | None = None,
        precision: Precision | str | None = None,
        processor_count: int | None = None,
        journal_file_names: None | str | list[str] = None,
        start_timeout: int = -1,
        additional_arguments: str = "",
        env: Dict[str, Any] | None = None,
        cleanup_on_exit: bool = True,
        start_transcript: bool = True,
        case_file_name: str | None = None,
        case_data_file_name: str | None = None,
        lightweight_mode: bool | None = None,
        py: bool | None = None,
        gpu: bool | None = None,
        cwd: str | None = None,
        fluent_path: str | None = None,
        topy: str | list | None = None,
        start_watchdog: bool | None = None,
        scheduler_options: dict | None = None,
        file_transfer_service: Any | None = None,
    ):
        """Launch Fluent session in standalone mode.

        Parameters
        ----------
        mode : FluentMode
            Launch mode of Fluent to point to a specific session type.
        ui_mode : UIMode
            Fluent user interface mode. Options are the values of the ``UIMode`` enum.
        graphics_driver : FluentWindowsGraphicsDriver or FluentLinuxGraphicsDriver
            Graphics driver of Fluent. Options are the values of the
            ``FluentWindowsGraphicsDriver`` enum in Windows or the values of the
            ``FluentLinuxGraphicsDriver`` enum in Linux.
        product_version : FluentVersion or str or float or int, optional
            Version of Ansys Fluent to launch. To use Fluent version 2025 R1, pass
            ``FluentVersion.v251``, ``"25.1.0"``, ``"25.1"``, ``25.1``, or ``251``.
            The default is ``None``, in which case the newest installed version is used.
        dimension : Dimension or int, optional
            Geometric dimensionality of the Fluent simulation. The default is ``None``,
            in which case ``Dimension.THREE`` is used. Options are either the values of the
            ``Dimension`` enum (``Dimension.TWO`` or ``Dimension.THREE``) or any of ``2`` and ``3``.
        precision : Precision or str, optional
            Floating point precision. The default is ``None``, in which case ``Precision.DOUBLE``
            is used. Options are either the values of the ``Precision`` enum (``Precision.SINGLE``
            or ``Precision.DOUBLE``) or any of ``"double"`` and ``"single"``.
        processor_count : int, optional
            Number of processors. The default is ``None``, in which case ``1``
            processor is used.  In job scheduler environments the total number of
            allocated cores is clamped to value of ``processor_count``.
        journal_file_names : str or list of str, optional
            The string path to a Fluent journal file or a list of such paths. Fluent executes the
            journals. The default is ``None``.
        start_timeout : int, optional
            Maximum allowable time in seconds for connecting to the Fluent
            server. The default is ``60``.
        additional_arguments : str, optional
            Additional arguments to send to Fluent as a string in the same
            format they are normally passed to Fluent on the command line.
        env : dict[str, str], optional
            Mapping to modify environment variables in Fluent. The default
            is ``None``.
        cleanup_on_exit : bool, optional
            Whether to shut down the connected Fluent session when PyFluent is
            exited, or the ``exit()`` method is called on the session instance,
            or if the session instance becomes unreferenced. The default is ``True``.
        start_transcript : bool, optional
            Whether to start streaming the Fluent transcript in the client. The
            default is ``True``. You can stop and start the streaming of the
            Fluent transcript subsequently via the method calls, ``transcript.start()``
            and ``transcript.stop()`` on the session object.
        case_file_name : str, optional
            Name of the case file to read into the
            Fluent session. The default is ``None``.
        case_data_file_name : str, optional
            Name of the case data file. If names of both a case file and case data file are provided, they are read into the Fluent session.
        lightweight_mode : bool, optional
            Whether to run in lightweight mode. In lightweight mode, the lightweight settings are read into the
            current Fluent solver session. The mesh is read into a background Fluent solver session,
            which replaces the current Fluent solver session once the mesh is read and the lightweight settings
            made by the user in the current Fluent solver session have been applied in the background Fluent
            solver session. This is all orchestrated by PyFluent and requires no special usage.
            This parameter is used only when ``case_file_name`` is provided. The default is ``False``.
        py : bool, optional
            If True, Fluent will run in Python mode. Default is None.
        gpu : bool, optional
            If True, Fluent will start with GPU Solver.
        cwd : str, Optional
            Working directory for the Fluent client.
        fluent_path: str, Optional
            User provided Fluent installation path.
        topy : bool or str, optional
            A boolean flag to write the equivalent Python journal(s) from the journal(s) passed.
            Can optionally take the file name of the new python journal file.
        start_watchdog : bool, optional
            When ``cleanup_on_exit`` is True, ``start_watchdog`` defaults to True,
            which means an independent watchdog process is run to ensure
            that any local GUI-less Fluent sessions started by PyFluent are properly closed (or killed if frozen)
            when the current Python process ends.
        scheduler_options : dict, optional
            Dictionary containing scheduler options. Default is None.

            Currently only the Slurm scheduler is supported. The ``scheduler_options``
            dictionary must be of the form ``{"scheduler": "slurm",
            "scheduler_headnode": "<headnode>", "scheduler_queue": "<queue>",
            "scheduler_account": "<account>"}``. The keys ``scheduler_headnode``,
            ``scheduler_queue`` and ``scheduler_account`` are optional and should be
            specified in a similar manner to Fluent's scheduler options.
        file_transfer_service : optional
            File transfer service for uploading and downloading files to and from the server.

        Returns
        -------
        :obj:`~typing.Union` [:class:`Meshing<ansys.fluent.core.session_meshing.Meshing>`, \
        :class:`~ansys.fluent.core.session_pure_meshing.PureMeshing`, \
        :class:`~ansys.fluent.core.session_solver.Solver`, \
        :class:`~ansys.fluent.core.session_solver_icing.SolverIcing`, dict]
            Session object or configuration dictionary if ``dry_run = True``.

        Notes
        -----
        Job scheduler environments such as SLURM, LSF, and PBS, allocate resources and compute nodes.
        The allocated machines and core counts are queried from the scheduler environment and
        passed to Fluent.
        """
        if not _SlurmWrapper.is_available():
            raise RuntimeError("Slurm is not available.")
        locals_ = locals().copy()
        argvals = {
            arg: locals_.get(arg)
            for arg in inspect.getargvalues(inspect.currentframe()).args
        }
        self._argvals, self._new_session = _get_argvals_and_session(argvals)
        self.file_transfer_service = file_transfer_service
        if os.getenv("PYFLUENT_SHOW_SERVER_GUI") == "1":
            ui_mode = UIMode.GUI
        self._argvals["ui_mode"] = UIMode(ui_mode)
        if self._argvals["start_timeout"] is None:
            self._argvals["start_timeout"] = -1
        if self._argvals["scheduler_options"]:
            if "scheduler" not in self._argvals["scheduler_options"]:
                raise InvalidArgument(
                    "The scheduler must be specified within scheduler options."
                )
            elif self._argvals["scheduler_options"]["scheduler"] != "slurm":
                raise InvalidArgument("Only slurm is supported as scheduler.")
            queue = self._argvals["scheduler_options"].get("scheduler_queue")
            if queue is not None:
                queues = _SlurmWrapper.list_queues()
                if queue not in queues:
                    raise InvalidArgument(
                        f"""Slurm queue is not valid. Valid queues are "{'", "'.join(queues)}"."""
                    )

    def _prepare(self):
        server_info_file_name_for_server, server_info_file_name_for_client = (
            _get_server_info_file_names(use_tmpdir=False)
        )
        self._server_info_file_name = server_info_file_name_for_client
        self._argvals.update(self._argvals["scheduler_options"])
        launch_cmd = _generate_launch_string(
            self._argvals,
            server_info_file_name_for_server,
        )

        self._sifile_last_mtime = Path(self._server_info_file_name).stat().st_mtime
        kwargs = _get_subprocess_kwargs_for_fluent(self._argvals["env"], self._argvals)
        launch_cmd += _build_journal_argument(
            self._argvals["topy"], self._argvals["journal_file_names"]
        )

        logger.debug(f"Launching Fluent with command: {launch_cmd}")
        proc = subprocess.Popen(launch_cmd, **kwargs)
        slurm_job_id = _get_slurm_job_id(proc)
        logger.info(f"Slurm job id = {slurm_job_id}")
        self._argvals["slurm_job_id"] = slurm_job_id
        return slurm_job_id

    def _launch(self, slurm_job_id) -> Meshing | PureMeshing | Solver | SolverIcing:
        _await_fluent_launch(
            self._server_info_file_name,
            self._argvals["start_timeout"],
            self._sifile_last_mtime,
        )
        session = self._new_session._create_from_server_info_file(
            server_info_file_name=self._server_info_file_name,
            file_transfer_service=None,
            cleanup_on_exit=self._argvals["cleanup_on_exit"],
            start_transcript=self._argvals["start_transcript"],
            inside_container=False,
        )
        return session

    def __call__(self) -> SlurmFuture:
        slurm_job_id = self._prepare()
        return SlurmFuture(
            ThreadPoolExecutor(max_workers=1).submit(self._launch, slurm_job_id),
            slurm_job_id,
        )



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module for transcript streaming."""

import os
from pathlib import Path

from ansys.api.fluent.v0 import transcript_pb2 as TranscriptModule
from ansys.fluent.core.streaming_services.streaming import StreamingService


class AppendToFile:
    """Class representing append to file action."""

    def __init__(self, file_name: str):
        """__init__ method of AppendToFile class."""
        self.f = open(file_name, "a")

    def __call__(self, transcript):
        self.f.write(transcript)
        self.f.flush()

    def __del__(self):
        self.f.close()


class Transcript(StreamingService):
    """Encapsulates a Fluent Transcript streaming service."""

    def __init__(self, transcript_service):
        """__init__ method of Transcript class."""
        super().__init__(
            stream_begin_method="BeginStreaming",
            target=Transcript._process_streaming,
            streaming_service=transcript_service,
        )
        self.callback_ids = []
        self._writing_transcript_to_interpreter = False

    def start(
        self, file_name: str | None = None, write_to_stdout: bool = False
    ) -> None:
        """Start streaming of Fluent transcript.

         Parameters
        ----------
        file_name: str, optional
            File path to write the transcript stream.
        write_to_stdout: bool, optional
            Flag to print transcript on the screen or not
        """
        if file_name:
            if Path(file_name).exists():
                os.remove(file_name)
            append_to_file = AppendToFile(file_name)
            self.callback_ids.append(
                self.register_callback(append_to_file, keep_new_lines=True)
            )
            if write_to_stdout:
                self._write_to_stdout()
        else:
            self._write_to_stdout()
        super().start()

    def _write_to_stdout(self):
        """Write transcript to stdout."""
        if not self._writing_transcript_to_interpreter:
            self.callback_ids.append(self.register_callback(print))
            self._writing_transcript_to_interpreter = True

    def stop(self) -> None:
        """Stop streaming of Fluent transcript."""
        for callback_id in self.callback_ids:
            self.unregister_callback(callback_id)
        super().stop()

    def _process_streaming(self, id, stream_begin_method, started_evt, *args, **kwargs):
        """Performs processes on transcript depending on the callback functions."""
        request = TranscriptModule.TranscriptRequest(*args, **kwargs)
        responses = self._streaming_service.begin_streaming(
            request, started_evt, id=id, stream_begin_method=stream_begin_method
        )
        transcript = ""
        while True:
            try:
                response = next(responses)
                with self._lock:
                    self._streaming = True
                    transcript += response.transcript
                    if transcript and transcript[-1] == "\n":
                        for callback_map in self._service_callbacks.values():
                            if "keep_new_lines" in callback_map[-1].keys():
                                if callback_map[-1]["keep_new_lines"]:
                                    callback_map[0](transcript)
                                else:
                                    callback_map[0](transcript[0:-1])
                            else:
                                callback_map[0](transcript[0:-1])
                        transcript = ""
            except StopIteration:
                break



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module for datamodel event streaming."""

import logging
import os
import threading
from typing import Callable

from google.protobuf.json_format import MessageToDict

from ansys.api.fluent.v0 import datamodel_se_pb2 as DataModelProtoModule
from ansys.fluent.core.services.datamodel_se import _convert_variant_to_value
from ansys.fluent.core.streaming_services.streaming import StreamingService

network_logger: logging.Logger = logging.getLogger("pyfluent.networking")


class DatamodelEvents(StreamingService):
    """Encapsulates a datamodel events streaming service."""

    def __init__(self, service):
        """Initialize DatamodelEvents."""
        super().__init__(
            stream_begin_method="BeginEventStreaming",
            target=DatamodelEvents._process_streaming,
            streaming_service=service,
        )
        self._cbs = {}
        service.event_streaming = self
        self._lock = threading.RLock()

    def register_callback(self, tag: str, cb: Callable):
        """Register a callback."""
        with self._lock:
            self._cbs[tag] = cb

    def unregister_callback(self, tag: str):
        """Unregister a callback."""
        with self._lock:
            self._cbs.pop(tag, None)

    def _process_streaming(self, id, stream_begin_method, started_evt, *args, **kwargs):
        """Processes datamodel events."""
        request = DataModelProtoModule.EventRequest(*args, **kwargs)
        responses = self._streaming_service.begin_streaming(
            request, started_evt, id=id, stream_begin_method=stream_begin_method
        )
        while True:
            try:
                response: DataModelProtoModule.EventResponse = next(responses)
                if os.getenv("PYFLUENT_HIDE_LOG_SECRETS") != "1":
                    network_logger.debug(
                        f"GRPC_TRACE: RPC = /grpcRemoting.DataModel/BeginEventStreaming, response = {MessageToDict(response)}"
                    )
                with self._lock:
                    self._streaming = True
                    cb = self._cbs.get(response.tag, None)
                    if cb:
                        if response.HasField("createdEventResponse"):
                            childtype = response.createdEventResponse.childtype
                            childname = response.createdEventResponse.childname
                            cb(childtype, childname)
                        elif response.HasField("attributeChangedEventResponse"):
                            value = response.attributeChangedEventResponse.value
                            cb(_convert_variant_to_value(value))
                        elif response.HasField("commandAttributeChangedEventResponse"):
                            value = response.commandAttributeChangedEventResponse.value
                            cb(_convert_variant_to_value(value))
                        elif response.HasField("modifiedEventResponse"):
                            value = response.modifiedEventResponse.value
                            cb(_convert_variant_to_value(value))
                        elif response.HasField("affectedEventResponse"):
                            cb()
                        elif response.HasField("deletedEventResponse"):
                            cb()
                        elif response.HasField("commandExecutedEventResponse"):
                            command = response.commandExecutedEventResponse.command
                            args = _convert_variant_to_value(
                                response.commandExecutedEventResponse.args
                            )
                            cb(command, args)
            except StopIteration:
                break



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import subprocess
import threading
import time

import pytest

import ansys.fluent.core as pyfluent
from ansys.fluent.core.examples import download_file
from ansys.fluent.core.fluent_connection import (
    WaitTypeError,
    _pid_exists,
    get_container,
)
from ansys.fluent.core.launcher.error_handler import IpPortNotProvided
from ansys.fluent.core.utils.execution import asynchronous, timeout_loop
from ansys.fluent.core.utils.fluent_version import FluentVersion
from docker.models.containers import Container


def _read_case(session, lightweight_setup=True):
    case_path = download_file("Static_Mixer_main.cas.h5", "pyfluent/static_mixer")
    # Ignore lightweight_setup variable for Fluent < 23.1 because not supported
    if session.get_fluent_version() < FluentVersion.v231:
        session.file.read(file_name=case_path, file_type="case")
    else:
        session.file.read(
            file_name=case_path, file_type="case", lightweight_setup=lightweight_setup
        )


def test_session_starts_transcript_by_default(new_solver_session) -> None:
    session = new_solver_session

    def print_transcript(transcript):
        print_transcript.called = True
        if transcript:
            print_transcript.transcript = transcript

    print_transcript.called = False
    print_transcript.transcript = None

    session.transcript.register_callback(print_transcript)

    _read_case(session=session)

    assert print_transcript.called
    assert print_transcript.transcript


def test_session_starts_no_transcript_if_disabled(
    new_solver_session,
) -> None:
    session = new_solver_session
    session.transcript.stop()

    def print_transcript(transcript):
        print_transcript.called = True

    print_transcript.called = False

    session.transcript.start(write_to_stdout=False)

    _read_case(session=session)

    assert not print_transcript.called


def test_server_exits_when_session_goes_out_of_scope() -> None:
    def f():
        session = pyfluent.launch_fluent()
        session.settings
        _fluent_host_pid = session.connection_properties.fluent_host_pid
        _cortex_host = session.connection_properties.cortex_host
        _inside_container = session.connection_properties.inside_container
        return _fluent_host_pid, _cortex_host, _inside_container

    fluent_host_pid, cortex_host, inside_container = f()

    timeout_loop(
        lambda: (inside_container and not get_container(cortex_host))
        or (not inside_container and not _pid_exists(fluent_host_pid)),
        60,
    )

    if inside_container:
        assert not get_container(cortex_host)
    else:
        assert not _pid_exists(fluent_host_pid)


def test_server_does_not_exit_when_session_goes_out_of_scope() -> None:
    def f():
        session = pyfluent.launch_fluent(cleanup_on_exit=False)
        session.settings
        _fluent_host_pid = session.connection_properties.fluent_host_pid
        _cortex_host = session.connection_properties.cortex_host
        _inside_container = session.connection_properties.inside_container
        _cortex_pwd = session.connection_properties.cortex_pwd
        return _fluent_host_pid, _cortex_host, _inside_container, _cortex_pwd

    fluent_host_pid, cortex_host, inside_container, cortex_pwd = f()
    time.sleep(10)
    if inside_container:
        assert get_container(cortex_host)
        subprocess.Popen(["docker", "stop", cortex_host])  # cortex_host = container_id
    else:
        from pathlib import Path

        assert _pid_exists(fluent_host_pid)
        if os.name == "nt":
            cleanup_file_ext = "bat"
            cmd_list = []
        elif os.name == "posix":
            cleanup_file_ext = "sh"
            cmd_list = ["bash"]
        else:
            raise Exception("Unrecognized operating system.")
        cleanup_file_name = (
            f"cleanup-fluent-{cortex_host}-{fluent_host_pid}.{cleanup_file_ext}"
        )
        print(f"cleanup_file_name: {cleanup_file_name}")
        cmd_list.append(Path(cortex_pwd, cleanup_file_name))
        print(f"cmd_list: {cmd_list}")
        subprocess.Popen(
            cmd_list,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )


def test_does_not_exit_fluent_by_default_when_connected_to_running_fluent(
    monkeypatch,
) -> None:
    session1 = pyfluent.launch_fluent()

    with pytest.raises(IpPortNotProvided):
        session2 = pyfluent.connect_to_fluent(
            ip=session1.connection_properties.ip,
            password=session1.connection_properties.password,
        )

    session2 = pyfluent.connect_to_fluent(
        ip=session1.connection_properties.ip,
        port=session1.connection_properties.port,
        password=session1.connection_properties.password,
    )
    assert session2.health_check.is_serving
    session2.exit()

    timeout_loop(
        session1.health_check.is_serving,
        5.0,
        expected="truthy",
    )

    assert session1.health_check.is_serving
    session1.exit()


def test_exit_fluent_when_connected_to_running_fluent(
    monkeypatch,
) -> None:  # import ansys.fluent.core as pyfluent
    session1 = pyfluent.launch_fluent(cleanup_on_exit=False)
    session2 = pyfluent.connect_to_fluent(
        ip=session1.connection_properties.ip,
        port=session1.connection_properties.port,
        password=session1.connection_properties.password,
        cleanup_on_exit=True,
    )
    session2.exit()

    timeout_loop(
        session1.health_check.is_serving,
        5.0,
        expected="falsy",
    )

    assert not session1.health_check.is_serving


def test_fluent_connection_properties(
    new_solver_session,
) -> None:
    connection_properties = new_solver_session.connection_properties
    assert isinstance(connection_properties.ip, str)
    assert isinstance(connection_properties.port, int)
    assert isinstance(connection_properties.password, str)
    assert isinstance(connection_properties.cortex_pwd, str)
    assert isinstance(connection_properties.cortex_pid, int)
    assert isinstance(connection_properties.cortex_host, str)
    assert isinstance(connection_properties.inside_container, (bool, Container))
    assert isinstance(connection_properties.fluent_host_pid, int)


def test_fluent_freeze_kill(
    new_solver_session,
) -> None:
    session = new_solver_session
    _read_case(session=session, lightweight_setup=False)

    def _freeze_fluent(s):
        try:
            s.tui.mesh.modify_zones.sep_face_zone_face("interior--fluid", "yes")
        except RuntimeError as expected_tui_error_from_force_exit:
            return expected_tui_error_from_force_exit
        return

    tmp_thread = threading.Thread(target=_freeze_fluent, args=(session,), daemon=True)
    tmp_thread.start()
    tmp_thread.join(5)
    if tmp_thread.is_alive():
        session.exit(timeout=1, timeout_force=True)
        tmp_thread.join()
    else:
        raise Exception("Test should have temporarily frozen Fluent, but did not.")

    assert session._fluent_connection.wait_process_finished(wait=5)


@pytest.mark.fluent_version(">=23.1")
def test_interrupt(static_mixer_case_session):
    solver = static_mixer_case_session
    solver.setup.general.solver.time = "unsteady-2nd-order"
    solver.solution.initialization.standard_initialize()
    asynchronous(solver.solution.run_calculation.dual_time_iterate)(
        time_step_count=100, max_iter_per_step=20
    )
    time.sleep(5)
    solver.solution.run_calculation.interrupt()
    assert solver.scheme_eval.scheme_eval("(rpgetvar 'time-step)") < 100


def test_fluent_exit(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.delenv("PYFLUENT_LOGGING")
    monkeypatch.delenv("PYFLUENT_WATCHDOG_DEBUG")
    inside_container = os.getenv("PYFLUENT_LAUNCH_CONTAINER")
    import ansys.fluent.core as pyfluent

    solver = pyfluent.launch_fluent(start_watchdog=False)
    cortex = (
        solver.connection_properties.cortex_host
        if inside_container
        else solver.connection_properties.cortex_pid
    )
    solver.exit()
    assert timeout_loop(
        lambda: (inside_container and not get_container(cortex))
        or (not inside_container and not _pid_exists(cortex)),
        timeout=60,
        idle_period=1,
    )


def test_fluent_exit_wait():
    session1 = pyfluent.launch_fluent()
    session1.exit()
    assert not session1._fluent_connection.wait_process_finished(wait=0)

    session2 = pyfluent.launch_fluent()
    session2.exit(wait=60)
    assert session2._fluent_connection.wait_process_finished(wait=0)

    session3 = pyfluent.launch_fluent()
    session3.exit(wait=True)
    assert session3._fluent_connection.wait_process_finished(wait=0)

    with pytest.raises(WaitTypeError):
        session4 = pyfluent.launch_fluent()
        session4.exit(wait="wait")



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""A package providing Fluent's Solver and Meshing capabilities in Python."""

import os
from pathlib import Path
import pydoc

# isort: off
# Logging has to be imported before importing other PyFluent modules
from ansys.fluent.core.logging import set_console_logging_level  # noqa: F401

# isort: on

from ansys.fluent.core.get_build_details import (  # noqa: F401
    get_build_version,
    get_build_version_string,
)
from ansys.fluent.core.launcher.launcher import (  # noqa: F401
    connect_to_fluent,
    launch_fluent,
)
from ansys.fluent.core.launcher.pyfluent_enums import (  # noqa: F401
    Dimension,
    FluentLinuxGraphicsDriver,
    FluentMode,
    FluentWindowsGraphicsDriver,
    Precision,
    UIMode,
)
from ansys.fluent.core.parametric import LocalParametricStudy  # noqa: F401
from ansys.fluent.core.pyfluent_warnings import (  # noqa: F401
    PyFluentDeprecationWarning,
    PyFluentUserWarning,
    warning,
)
from ansys.fluent.core.search import search  # noqa: F401
from ansys.fluent.core.services.batch_ops import BatchOps  # noqa: F401
from ansys.fluent.core.session import BaseSession as Fluent  # noqa: F401
from ansys.fluent.core.streaming_services.events_streaming import *  # noqa: F401, F403
from ansys.fluent.core.utils import fldoc, get_examples_download_dir
from ansys.fluent.core.utils.fluent_version import FluentVersion  # noqa: F401
from ansys.fluent.core.utils.setup_for_fluent import setup_for_fluent  # noqa: F401

__version__ = "0.30.5"

_VERSION_INFO = "Build date: April 29, 2025 16:52 UTC ShaID: f730b88"
"""
Global variable indicating the version info of the PyFluent package.
Build timestamp and commit hash are added to this variable during packaging.
"""

_THIS_DIRNAME = os.path.dirname(__file__)
_README_FILE = os.path.normpath(os.path.join(_THIS_DIRNAME, "docs", "README.rst"))

if os.path.exists(_README_FILE):
    with open(_README_FILE, encoding="utf8") as f:
        __doc__ = f.read()


def version_info() -> str:
    """Method returning the version of PyFluent being used.

    Returns
    -------
    str
        The PyFluent version being used.

    Notes
    -------
    Only available in packaged versions. Otherwise it will return __version__.
    """
    return _VERSION_INFO if _VERSION_INFO is not None else __version__


# Latest released Fluent version
FLUENT_RELEASE_VERSION = "25.1.0"

# Current dev Fluent version
FLUENT_DEV_VERSION = "25.2.0"

# Path to the example input/data files are downloaded
EXAMPLES_PATH = str(get_examples_download_dir())

# Host path which is mounted to the container
CONTAINER_MOUNT_SOURCE = None

# Path inside the container where the host path is mounted
CONTAINER_MOUNT_TARGET = "/mnt/pyfluent"

# Set this to False to stop automatically inferring and setting REMOTING_SERVER_ADDRESS
INFER_REMOTING_IP = True

# Time in second to wait for response for each ip while inferring remoting ip
INFER_REMOTING_IP_TIMEOUT_PER_IP = 2

pydoc.text.docother = fldoc.docother.__get__(pydoc.text, pydoc.TextDoc)

# Whether to use datamodel state caching
DATAMODEL_USE_STATE_CACHE = True

# Whether to use datamodel attribute caching
DATAMODEL_USE_ATTR_CACHE = True

# Whether to stream and cache commands state
DATAMODEL_USE_NOCOMMANDS_DIFF_STATE = True

# Whether to return the state changes on mutating datamodel rpcs
DATAMODEL_RETURN_STATE_CHANGES = True

# Whether to use remote gRPC file transfer service
USE_FILE_TRANSFER_SERVICE = False

# Directory where API files are written out during codegen
CODEGEN_OUTDIR = os.getenv(
    "PYFLUENT_CODEGEN_OUTDIR", (Path(__file__) / ".." / "generated").resolve()
)

# Whether to show mesh in Fluent after case read
FLUENT_SHOW_MESH_AFTER_CASE_READ = False

# Whether to write the automatic transcript in Fluent
FLUENT_AUTOMATIC_TRANSCRIPT = False

# Whether to interrupt Fluent solver from PyFluent
SUPPORT_SOLVER_INTERRUPT = False

# Whether to start watchdog
START_WATCHDOG = None

# Whether to skip health check
CHECK_HEALTH = True

# Whether to print search results
PRINT_SEARCH_RESULTS = True

# Whether to clear environment variables related to Fluent parallel mode
CLEAR_FLUENT_PARA_ENVS = False

# Set stdout of the launched Fluent process
# Valid values are same as subprocess.Popen's stdout argument
LAUNCH_FLUENT_STDOUT = None

# Set stderr of the launched Fluent process
# Valid values are same as subprocess.Popen's stderr argument
LAUNCH_FLUENT_STDERR = None

# Set the IP address of the Fluent server while launching Fluent
LAUNCH_FLUENT_IP = None

# Set the port of the Fluent server while launching Fluent
LAUNCH_FLUENT_PORT = None

# Skip password check during rpc execution when Fluent is launched from PyFluent
LAUNCH_FLUENT_SKIP_PASSWORD_CHECK = False

# Set Fluent precision mode
FLUENT_PRECISION_MODE = None



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _ref_radiation_headlamp:

Modeling Radiation in a Headlamp Using the Monte Carlo Method
-------------------------------------------------------------
This example solves for the radiative and conductive heat transfer within a car
headlamp exposed to the sun's rays to determine the severity of any hotspots
that form. It uses a Monte Carlo radiation model and the  pressure-based
solver. This is based on the Fluent tutorial titled "Using the Monte Carlo
Radiation Model".

**Workflow tasks**

The Modeling Radiation Using the Monte Carlo Method example guides you through
these tasks:

- Creation of a mesh using the Watertight Geometry workflow.
- Setting up a Monte Carlo radiation model.
- Creation of materials with thermal and radiation properties.
- Setting boundary conditions for heat transfer and radiation calculations.
- Calculating a solution using the pressure-based solver.

**Problem description**

The problem considers the headlamp of a parked car exposed to sunlight. The
lens focuses incoming radiation onto the internal components of the headlamp,
producing thermal hotspots that could cause damage due to thermal stresses or
material degradation.
"""

###############################################################################
# .. image:: ../../_static/radiation_headlamp_geom.png
#   :width: 500pt
#   :align: center

###############################################################################
# Example Setup
# -------------
# Before you can use the watertight geometry meshing workflow, you must set up
# the example and initialize this workflow.
#
# Perform required imports
# ~~~~~~~~~~~~~~~~~~~~~~~~
# Perform required imports, which includes downloading and importing the
# geometry files.

import ansys.fluent.core as pyfluent
from ansys.fluent.core import examples

headlamp_spaceclaim_file, headlamp_pmdb_file = [
    examples.download_file(
        f,
        "pyfluent/radiation_headlamp",
    )
    for f in ["headlamp.scdoc", "headlamp.pmdb"]
]

###############################################################################
# Launch Fluent
# ~~~~~~~~~~~~~
# Launch Fluent as a service in meshing mode with single precision running on
# four processors and print Fluent version.

meshing = pyfluent.launch_fluent(
    precision="single",
    processor_count=4,
    mode="meshing",
)
print(meshing.get_fluent_version())

###############################################################################
# Initialize workflow
# ~~~~~~~~~~~~~~~~~~~
# Initialize the watertight geometry meshing workflow.

meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")

###############################################################################
# Watertight geometry meshing workflow
# ------------------------------------
# The fault-tolerant meshing workflow guides you through the several tasks that
# follow.
#
# Import CAD and set length units
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Import the CAD geometry and set the length units to millimeters.

geo_import = meshing.workflow.TaskObject["Import Geometry"]
geo_import.Arguments.set_state(
    {
        "FileName": headlamp_pmdb_file,
        "LengthUnit": "mm",
    }
)

geo_import.Execute()

###############################################################################
# Add local sizing
# ~~~~~~~~~~~~~~~~
# Add local sizing controls to the geometry.

local_sizing = meshing.workflow.TaskObject["Add Local Sizing"]
local_sizing.Arguments.set_state(
    {
        "AddChild": "yes",
        "BOIControlName": "boi_lens",
        "BOIExecution": "Body Of Influence",
        "BOIFaceLabelList": ["boi"],
        "BOISize": 2,
    }
)

local_sizing.AddChildAndUpdate()

local_sizing.Arguments.set_state(
    {
        "AddChild": "yes",
        "BOIControlName": "bodysize_lens",
        "BOIExecution": "Body Size",
        "BOIFaceLabelList": ["lens"],
        "BOISize": 2,
    }
)

local_sizing.AddChildAndUpdate()

###############################################################################
# Generate surface mesh
# ~~~~~~~~~~~~~~~~~~~~~
# Generate the surface mesh.

surface_mesh_gen = meshing.workflow.TaskObject["Generate the Surface Mesh"]
surface_mesh_gen.Arguments.set_state(
    {
        "CFDSurfaceMeshControls": {
            "MinSize": 1,
            "MaxSize": 40,
        }
    }
)

surface_mesh_gen.Execute()

###############################################################################
# Improve surface mesh
# ~~~~~~~~~~~~~~~~~~~~
# Improve the surface mesh.

surface_mesh_gen.InsertNextTask(CommandName="ImproveSurfaceMesh")

meshing.workflow.TaskObject["Improve Surface Mesh"].Execute()

###############################################################################
# Describe geometry
# ~~~~~~~~~~~~~~~~~
# Describe geometry and define the fluid region.

describe_geo = meshing.workflow.TaskObject["Describe Geometry"]
describe_geo.Arguments.set_state(
    {
        "SetupType": "The geometry consists of both fluid and solid regions and/or voids",
        "CappingRequired": "No",
        "WallToInternal": "No",
        "InvokeShareTopology": "No",
        "Multizone": "No",
    }
)

describe_geo.Execute()

###############################################################################
# Update boundaries
# ~~~~~~~~~~~~~~~~~
# Update the boundaries.

update_bc = meshing.workflow.TaskObject["Update Boundaries"]
update_bc.Arguments.set_state(
    {
        "BoundaryLabelList": ["rad-input"],
        "BoundaryLabelTypeList": ["wall"],
    }
)

update_bc.Execute()

###############################################################################
# Create fluid region
# ~~~~~~~~~~~~~~~~~~~
# Create the fluid region.

create_regions = meshing.workflow.TaskObject["Create Regions"]
create_regions.Arguments.set_state({"NumberOfFlowVolumes": 1})

create_regions.Execute()

###############################################################################
# Update regions
# ~~~~~~~~~~~~~~
# Update the regions.

meshing.workflow.TaskObject["Update Regions"].Execute()

###############################################################################
# Boundary layers
# ~~~~~~~~~~~~~~~~~~~
# Do not add boundary layers and proceed to the next task.

add_boundary_layers = meshing.workflow.TaskObject["Add Boundary Layers"]
add_boundary_layers.Arguments.set_state({"AddChild": "no"})

add_boundary_layers.Execute()

###############################################################################
# Generate volume mesh
# ~~~~~~~~~~~~~~~~~~~~
# Generate the volume mesh, which consists of setting properties for the
# volume mesh.

volume_mesh_gen = meshing.workflow.TaskObject["Generate the Volume Mesh"]
volume_mesh_gen.Arguments.set_state(
    {
        "VolumeMeshPreferences": {
            "PolyFeatureAngle": 40,
        },
    },
)
volume_mesh_gen.Execute()

###############################################################################
# Check mesh in meshing mode
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
# Check the mesh in meshing mode.

meshing.tui.mesh.check_mesh()

###############################################################################
# Save mesh file
# ~~~~~~~~~~~~~~
# Save the mesh file (``headlamp.msh.h5``).

meshing.meshing.File.WriteMesh(FileName="headlamp.msh.h5")

###############################################################################
# Solve and postprocess
# ---------------------
# Once you have completed the watertight geometry meshing workflow, you can
# solve and postprocess the results.
#
# Switch to solution mode
# ~~~~~~~~~~~~~~~~~~~~~~~
# Switch to solution mode. Now that a high-quality mesh has been generated
# using Fluent in meshing mode, you can switch to solver mode to complete the
# setup of the simulation.

solver = meshing.switch_to_solver()

###############################################################################
# Enable energy and viscosity models
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Set up the energy and viscosity models.

models = solver.settings.setup.models
models.energy.enabled = True
models.viscous.model = "laminar"

###############################################################################
# Set up radiation model
# ~~~~~~~~~~~~~~~~~~~~~~
# Set up the Monte Carlo radiation model. The number of histories is set to 10
# million in order to reduce computation time, but this may need to be
# increased to obtain accurate results.

radiation = models.radiation
radiation.model = "monte-carlo"
radiation.monte_carlo.number_of_histories = 1e7
radiation.solve_frequency.iteration_interval = 20

###############################################################################
# Define materials
# ~~~~~~~~~~~~~~~~
# Create materials to represent the glass and plastic parts of the headlamp.
# To demonstrate two different methods of creating materials through the
# settings API, we will create glass using a dictionary and plastic using dot
# syntax.

# --- Properties of glass ---
# Density: 2650 [kg/m^3]
# Specific heat capacity: 1887 [J/(kg K)]
# Thermal conductivity: 7.6 [W/(m K)]
# Absorption coefficient: 5.302
# Refractive index: 1.4714

glass = solver.settings.setup.materials.solid.create("glass")
glass.set_state(
    {
        "chemical_formula": "",
        "density": {
            "option": "constant",
            "value": 2650,
        },
        "specific_heat": {
            "option": "constant",
            "value": 1887,
        },
        "thermal_conductivity": {
            "option": "constant",
            "value": 7.6,
        },
        "absorption_coefficient": {
            "option": "constant",
            "value": 5.302,
        },
        "refractive_index": {
            "option": "constant",
            "value": 1.4714,
        },
    }
)

# --- Properties of plastic ---
# Density: 1545.3 [kg/m^3]
# Specific heat capacity: 2302 [J/(kg K)]
# Thermal conductivity: 0.316 [W/(m K)]

plastic = solver.settings.setup.materials.solid.create("plastic")
plastic.chemical_formula = ""
plastic.density.value = 1545.3
plastic.specific_heat.value = 2302
plastic.thermal_conductivity.value = 0.316
plastic.absorption_coefficient.value = 0
plastic.refractive_index.value = 1

###############################################################################
# Cell Zone Conditions
# ~~~~~~~~~~~~~~~~~~~~
# Set the cell zone conditions for the bezel and the lens.

solver.settings.setup.cell_zone_conditions.solid["bezel"].material = "plastic"
solver.settings.setup.cell_zone_conditions.copy(
    from_="bezel",
    to=[
        "holder",
        "housing",
        "inner-bezel",
        "reflector",
        "rim-bezel",
        "seating-steel-rim",
    ],
)

lens_cellzone_conds = solver.settings.setup.cell_zone_conditions.solid["lens"]
lens_cellzone_conds.material = "glass"
lens_cellzone_conds.general.participates_in_radiation = True

###########################################################################################################
# Boundary Conditions
# ~~~~~~~~~~~~~~~~~~~
# Set the boundary conditions.

# --- Set up bezel-enclosure BC ---
# Material: plastic
# BC type: opaque
# Internal emissivity: 1
# Diffuse fraction: 1

bezel_enc_bc = solver.settings.setup.boundary_conditions.wall["bezel-enclosure"]
bezel_enc_bc.thermal.material = "plastic"
bezel_enc_bc.radiation.radiation_bc = "Opaque"
bezel_enc_bc.radiation.internal_emissivity = 1
bezel_enc_bc.radiation.diffuse_irradiation_settings.diffuse_fraction_band = {"s-": 1}

# Get list of wall zones
bc_state = solver.settings.setup.boundary_conditions.get_state()
zones_to_copy = list(bc_state["wall"].keys())
incompatible_zones = ["bezel-enclosure", "enclosure:1", "rad-input"]
for incompatible_zone in incompatible_zones:
    zones_to_copy.remove(incompatible_zone)

# Copy bezel-enclosure BC to all other BCs
solver.settings.setup.boundary_conditions.copy(
    from_="bezel-enclosure",
    to=zones_to_copy,
)

# --- Set up enclosure-lens BC ---
# Material: glass
# BC type: semi-transparent
# Diffuse fraction: 0

enc_lens_bc = solver.settings.setup.boundary_conditions.wall["enclosure-lens"]
enc_lens_bc.thermal.material = "glass"
enc_lens_bc.radiation.radiation_bc = "Semi Transparent"
enc_lens_bc.radiation.diffuse_irradiation_settings.diffuse_fraction_band = {"s-": 0}

# Copy enclosure-lens BC to other lens boundary
solver.settings.setup.boundary_conditions.copy(
    from_="enclosure-lens",
    to=["enclosure-lens-shadow"],
)

# --- Set up enclosure-rim-bezel BC ---
# Material: plastic
# BC type: opaque
# Internal emissivity: 0.16
# Diffuse fraction: 0.1

enc_rim_bezel_bc = solver.settings.setup.boundary_conditions.wall["enclosure-rim-bezel"]
enc_rim_bezel_bc.thermal.material = "plastic"
enc_rim_bezel_bc.radiation.radiation_bc = "Opaque"
enc_rim_bezel_bc.radiation.internal_emissivity = 0.16
enc_rim_bezel_bc.radiation.diffuse_irradiation_settings.diffuse_fraction_band = {
    "s-": 0.1
}

# Copy enclosure-rim-bezel BC to other rim bezel boundaries
solver.settings.setup.boundary_conditions.copy(
    from_="enclosure-rim-bezel",
    to=[
        "enclosure-rim-bezel-shadow",
        "holder-rim-bezel",
        "holder-rim-bezel-shadow",
        "housing-rim-bezel",
        "housing-rim-bezel-shadow",
    ],
)

# --- Set up enclosure:1 (domain boundaries) BC ---
# BC type: temperature
# Temperature: 298.15 [K]

enc1_bc = solver.settings.setup.boundary_conditions.wall["enclosure:1"]
enc1_bc.thermal.thermal_condition = "Temperature"
enc1_bc.thermal.temperature.value = 298.15

# --- Set up radiation input BC ---
# BC type: temperature
# Temperature: 298.15 [K]
# Boundary source: yes
# Direct irradiation: 1200 [W/m^2]
# Radiation direction: (-0.848, 0, -0.53)

rad_input_bc = solver.settings.setup.boundary_conditions.wall["rad-input"]
rad_input_bc.thermal.thermal_condition = "Temperature"
rad_input_bc.thermal.temperature.value = 298.15
rad_input_bc.radiation.boundary_source = True
rad_input_bc.radiation.direct_irradiation_settings.direct_irradiation = {"s-": 1200}
rad_input_bc.radiation.direct_irradiation_settings.reference_direction = [
    -0.848,
    0,
    -0.53,
]

###########################################################################################################
# Set convergence criteria
# ~~~~~~~~~~~~~~~~~~~~~~~~
# Enable residual plots and set the convergence criteria to 'none'.

solver.solution.monitor.residual.options.criterion_type = "none"

###########################################################################################################
# Define surface reports
# ~~~~~~~~~~~~~~~~~~~~~~
# Define a surface report to find the maximum temperature of the inner bezel,
# then print the state of the report object.

solver.settings.solution.report_definitions.surface["max-temp"] = {}
max_temp_surf_report = solver.settings.solution.report_definitions.surface["max-temp"]
max_temp_surf_report.surface_names = ["enclosure-inner-bezel"]
max_temp_surf_report.report_type = "surface-facetmax"
max_temp_surf_report.field = "temperature"

solver.settings.solution.monitor.report_plots["max-temp-rplot"] = {}
max_temp_surf_report_plot = solver.settings.solution.monitor.report_plots[
    "max-temp-rplot"
]
max_temp_surf_report_plot.report_defs = "max-temp"
max_temp_surf_report_plot.print = True

###########################################################################################################
# Define report plots
# ~~~~~~~~~~~~~~~~~~~
# Define a plot of the maximum temperature.

solver.settings.solution.monitor.report_plots["max-temp-rplot"] = {}
max_temp_rplot = solver.settings.solution.monitor.report_plots["max-temp-rplot"]
max_temp_rplot.report_defs = "max-temp"
max_temp_rplot.print = True

###############################################################################
# Save case file
# ~~~~~~~~~~~~~~
# Save the case file (``headlamp.cas.h5``).

solver.settings.file.write(file_name="headlamp.cas.h5", file_type="case")

###############################################################################
# Initialize flow field
# ~~~~~~~~~~~~~~~~~~~~~
# Initialize the solution.

solver.settings.solution.initialization.initialize()

###############################################################################
# Solve for 19 iterations
# ~~~~~~~~~~~~~~~~~~~~~~~
# Solve for 19 iterations. 99 iterations is recommended by the tutorial, but is
# reduced to 19 for this example for demonstration purposes.

solver.settings.solution.run_calculation.iterate(iter_count=19)

###############################################################################
# Write final case file and data
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Enable overwrite so that the original case file will be overwritten. Write
# the final case file and the data.

solver.settings.file.batch_options.confirm_overwrite = True
solver.settings.file.write(file_name="headlamp.cas.h5", file_type="case-data")

###############################################################################
# Close Fluent
# ~~~~~~~~~~~~
# Close Fluent.

solver.exit()

###############################################################################



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Base classes for builtin setting classes."""

from typing import Protocol, runtime_checkable

from ansys.fluent.core.solver.flobject import NamedObject, SettingsBase
from ansys.fluent.core.solver.settings_builtin_data import DATA
from ansys.fluent.core.utils.fluent_version import FluentVersion


@runtime_checkable
class Solver(Protocol):
    """Solver session class for type hinting."""

    settings: SettingsBase


def _get_settings_root(settings_source: SettingsBase | Solver):
    def is_root_obj(obj):
        return isinstance(obj, SettingsBase) and obj.parent is None

    if is_root_obj(settings_source):
        return settings_source
    elif isinstance(settings_source, Solver) and is_root_obj(settings_source.settings):
        return settings_source.settings
    else:
        raise TypeError(
            f"{settings_source} is not a Settings root or a Solver session object."
        )


def _get_settings_obj(settings_root, builtin_settings_obj):
    builtin_cls_name = builtin_settings_obj.__class__.__name__
    obj = settings_root
    path = DATA[builtin_cls_name][1]
    if isinstance(path, dict):
        version = FluentVersion(obj._version)
        path = path.get(version)
        if path is None:
            raise RuntimeError(
                f"{builtin_cls_name} is not supported in Fluent version {version}."
            )
    comps = path.split(".")
    for i, comp in enumerate(comps):
        obj = SettingsBase.__getattribute__(obj, comp)  # bypass InactiveObjectError
        if i < len(comps) - 1 and isinstance(obj, NamedObject):
            obj_name = getattr(builtin_settings_obj, comp)
            obj = obj[obj_name]
    return obj


class _SingletonSetting:
    # Covers both groups and named-object containers
    def __init__(self, settings_source: SettingsBase | Solver | None = None, **kwargs):
        self.__dict__.update(dict(settings_source=None) | kwargs)
        if settings_source is not None:
            self.settings_source = settings_source

    def __setattr__(self, name, value):
        if name == "settings_source":
            settings_root = _get_settings_root(value)
            obj = _get_settings_obj(settings_root, self)
            self.__class__ = obj.__class__
            self.__dict__.clear()
            self.__dict__.update(obj.__dict__ | dict(settings_source=settings_root))
        else:
            super().__setattr__(name, value)


class _NonCreatableNamedObjectSetting:
    def __init__(
        self, name: str, settings_source: SettingsBase | Solver | None = None, **kwargs
    ):
        self.__dict__.update(dict(settings_source=None, name=name) | kwargs)
        if settings_source is not None:
            self.settings_source = settings_source

    def __setattr__(self, name, value):
        if name == "settings_source":
            settings_root = _get_settings_root(value)
            obj = _get_settings_obj(settings_root, self)
            obj = obj[self.name]
            self.__class__ = obj.__class__
            self.__dict__.clear()
            self.__dict__.update(obj.__dict__ | dict(settings_source=settings_root))
        else:
            super().__setattr__(name, value)


class _CreatableNamedObjectSetting:
    def __init__(
        self,
        settings_source: SettingsBase | Solver | None = None,
        name: str | None = None,
        new_instance_name: str | None = None,
        **kwargs,
    ):
        if name and new_instance_name:
            raise ValueError("Cannot specify both name and new_instance_name.")
        self.__dict__.update(
            dict(settings_source=None, name=name, new_instance_name=new_instance_name)
            | kwargs
        )
        if settings_source is not None:
            self.settings_source = settings_source

    def __setattr__(self, name, value):
        if name == "settings_source":
            settings_root = _get_settings_root(value)
            obj = _get_settings_obj(settings_root, self)
            if self.name:
                obj = obj[self.name]
            elif self.new_instance_name:
                obj = obj.create(self.new_instance_name)
            else:
                obj = obj.create()
            self.__class__ = obj.__class__
            self.__dict__.clear()
            self.__dict__.update(obj.__dict__ | dict(settings_source=settings_root))
        else:
            super().__setattr__(name, value)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
#################################
Automotive Brake Thermal Analysis
#################################

Objective:
==========

Braking surfaces get heated due to frictional heating during braking.
High temperature affects the braking performance and life of the braking system.
This example demonstrates:

* Fluent setup and simulation using PyFluent
* Post processing using PyVista (3D Viewer) and Matplotlib (2D graphs)

"""

####################################################################################
# Import required libraries/modules
# ==================================================================================

import csv

import matplotlib.pyplot as plt

import ansys.fluent.core as pyfluent
from ansys.fluent.core import examples
import ansys.fluent.visualization.pyvista as pv

import_filename = examples.download_file(
    "brake.msh.h5",
    "pyfluent/examples/Brake-Thermal-PyVista-Matplotlib",
)

####################################################################################
# Fluent Solution Setup
# ==================================================================================

####################################################################################
# Launch Fluent session with solver mode and print Fluent version
# ---------------------------------------------------------------

session = pyfluent.launch_fluent(precision="double", processor_count=2, version="3d")
print(session.get_fluent_version())

####################################################################################
# Import mesh
# ------------

session.settings.file.read_case(file_name=import_filename)

############################
# Define models and material
# --------------------------
session.settings.setup.models.energy = {"enabled": True}
session.settings.setup.general.solver.time = "unsteady-2nd-order-bounded"
session.tui.define.materials.copy("solid", "steel")

#########################################
# Solve only energy equation (conduction)
# ---------------------------------------
session.tui.solve.set.equations("flow", "no", "kw", "no")

##########################################################################################
# Define disc rotation
# --------------------
# (15.79 rps corresponds to 100 km/h car speed
# with 0.28 m of axis height from ground)
session.settings.setup.cell_zone_conditions.solid["disc2"] = {
    "solid_motion": {
        "solid_motion_zone_motion_function": "none",
        "solid_motion_axis_direction": [0, 1, 0],
        "solid_motion_axis_origin": [-0.035, -0.821, 0.045],
        "solid_motion_velocity": [0, 0, 0],
        "solid_omega": -15.79,
        "solid_relative_to_thread": "absolute",
        "enable": True,
    }
}
session.settings.setup.cell_zone_conditions.solid["disc1"] = {
    "solid_motion": {
        "solid_motion_zone_motion_function": "none",
        "solid_motion_axis_direction": [0, 1, 0],
        "solid_motion_axis_origin": [-0.035, -0.821, 0.045],
        "solid_motion_velocity": [0, 0, 0],
        "solid_omega": -15.79,
        "solid_relative_to_thread": "absolute",
        "enable": True,
    }
}

##########################################################################################
# Apply frictional heating on pad-disc surfaces
# ----------------------------------------------
# Wall thickness 0f 2 mm has been assumed and 2e9 w/m3 is the heat generation which
# has been calculated from kinetic energy change due to braking.

session.settings.setup.boundary_conditions.wall["wall-pad-disc2"] = {
    "thermal": {"q_dot": {"value": 2000000000}, "wall_thickness": {"value": 0.002}}
}
session.settings.setup.boundary_conditions.wall["wall_pad-disc1"] = {
    "thermal": {"q_dot": {"value": 2000000000}, "wall_thickness": {"value": 0.002}}
}

##########################################################################################
# Apply convection cooling on outer surfaces due to air flow
# -----------------------------------------------------------
# Outer surfaces are applied a constant htc of 100 W/(m2 K)
# and 300 K free stream temperature

session.tui.define.boundary_conditions.set.wall(
    "wall-disc*",
    "wall-geom*",
    "()",
    "thermal-bc",
    "yes",
    "convection",
    "convective-heat-transfer-coefficient",
    "no",
    100,
    "q",
)

##########################################################################################
# Initialize
# ----------
# Initialize with 300 K temperature

session.settings.solution.initialization.initialization_type = "standard"
session.settings.solution.initialization.standard_initialize()

###############################################################################################
# Post processing setup
# ---------------------
# * Report definitions and monitor plots
# * Set contour plot properties
# * Set views and camera
# * Set animation object

session.settings.solution.report_definitions.volume["max-pad-temperature"] = {}
session.settings.solution.report_definitions.volume[
    "max-pad-temperature"
].report_type = "volume-max"
session.settings.solution.report_definitions.volume["max-pad-temperature"] = {
    "field": "temperature",
    "cell_zones": ["geom-1-innerpad", "geom-1-outerpad"],
}

session.settings.solution.report_definitions.volume["max-disc-temperature"] = {}
session.settings.solution.report_definitions.volume[
    "max-disc-temperature"
].report_type = "volume-max"
session.settings.solution.report_definitions.volume["max-disc-temperature"] = {
    "field": "temperature",
    "cell_zones": ["disc1", "disc2"],
}

session.settings.solution.monitor.report_plots.create(name="max-temperature")
session.settings.solution.monitor.report_plots["max-temperature"] = {
    "report_defs": ["max-pad-temperature", "max-disc-temperature"]
}

session.settings.solution.monitor.report_files.create(name="max-temperature")
session.settings.solution.monitor.report_files["max-temperature"] = {
    "report_defs": ["max-pad-temperature", "max-disc-temperature"],
    "file_name": "max-temperature.out",
}
session.settings.solution.monitor.report_files["max-temperature"].report_defs = [
    "max-pad-temperature",
    "max-disc-temperature",
    "flow-time",
]

session.settings.results.graphics.contour.create(name="contour-1")
session.settings.results.graphics.contour["contour-1"] = {
    "boundary_values": True,
    "range_option": {"auto_range_on": {"global_range": True}},
    "field": "temperature",
    "draw_mesh": False,
    "coloring": {"smooth": False},
    "color_map": {
        "user_skip": 9,
        "log_scale": False,
        "visible": True,
        "width": 6,
        "show_all": True,
        "font_name": "Helvetica",
        "font_size": 0.032,
        "font_automatic": True,
        "length": 0.54,
        "size": 100,
        "format": "%0.2e",
        "position": 1,
        "color": "field-velocity",
    },
    "mesh_object": "",
    "node_values": True,
    "contour_lines": False,
    "display_state_name": "None",
    "filled": True,
}

session.settings.results.graphics.contour["temperature"] = {
    "field": "temperature",
    "surfaces_list": "wall*",
    "color_map": {
        "visible": True,
        "size": 100,
        "color": "field-velocity",
        "log_scale": False,
        "format": "%0.1f",
        "user_skip": 9,
        "show_all": True,
        "position": 1,
        "font_name": "Helvetica",
        "font_automatic": True,
        "font_size": 0.032,
        "length": 0.54,
        "width": 6,
        "bground_transparent": True,
        "bground_color": "#CCD3E2",
        "title_elements": "Variable and Object Name",
    },
}

session.settings.results.graphics.contour["temperature"].range_option.option = (
    "auto-range-off"
)
session.settings.results.graphics.contour["temperature"].range_option.set_state(
    {
        "auto_range_off": {"maximum": 400.0, "minimum": 300, "clip_to_range": False},
    }
)

session.settings.results.graphics.views.restore_view(view_name="top")
session.settings.results.graphics.views.camera.zoom(factor=2)
session.settings.results.graphics.views.save_view(view_name="animation-view")

session.settings.solution.calculation_activity.solution_animations[
    "animate-temperature"
] = {
    "animate_on": "temperature",
    "frequency_of": "flow-time",
    "flow_time_frequency": 0.05,
    "view": "animation-view",
}

##################################################################################################
# Run simulation
# ---------------
# * Run simulation for 2 seconds flow time
# * Set time step size
# * Set number of time steps and maximum number of iterations per time step

session.settings.solution.run_calculation.transient_controls.time_step_size = 0.01
session.settings.solution.run_calculation.dual_time_iterate(
    time_step_count=200, max_iter_per_step=5
)

##################################################################################################
# Save simulation data
# --------------------
# Write case and data files
session.settings.file.write(file_type="case-data", file_name="brake-final.cas.h5")

###############################################
# Post processing with PyVista (3D visualization)
# ===============================================

###############################################
# Create a graphics session
# -------------------------
graphics_session1 = pv.Graphics(session)

###############################################
# Temperature contour object
# --------------------------
contour1 = graphics_session1.Contours["temperature"]

###############################################
# Check available options for contour object
# -------------------------------------------

contour1()

###############################################
# Set contour properties
# ----------------------

contour1.field = "temperature"
contour1.surfaces_list = [
    "wall-disc1",
    "wall-disc2",
    "wall-pad-disc2",
    "wall_pad-disc1",
    "wall-geom-1-bp_inner",
    "wall-geom-1-bp_outer",
    "wall-geom-1-innerpad",
    "wall-geom-1-outerpad",
]
contour1.range.option = "auto-range-off"
contour1()
contour1.range.auto_range_off.minimum = 300
contour1.range.auto_range_off.maximum = 400

###############################################
# Display contour
# ---------------

contour1.display()

# %%
# .. image:: ../../_static/brake_surface_temperature.png
#    :align: center
#    :alt: Brake Surface Temperature Contour

# %%
#    Brake Surface Temperature

###############################################
# Post processing with Matplotlib (2D graph)
# ===============================================

###############################################
# Read monitor file
# -----------------

X = []
Y = []
Z = []
i = -1
with open("max-temperature.out", "r") as datafile:
    plotting = csv.reader(datafile, delimiter=" ")
    for rows in plotting:
        i = i + 1
        if i > 2:
            X.append(float(rows[3]))
            Y.append(float(rows[2]))
            Z.append(float(rows[1]))

###############################################
# Plot graph
# ----------


plt.title("Maximum Temperature", fontdict={"color": "darkred", "size": 20})
plt.plot(X, Z, label="Max. Pad Temperature", color="red")
plt.plot(X, Y, label="Max. Disc Temperature", color="blue")
plt.xlabel("Time (sec)")
plt.ylabel("Max Temperature (K)")
plt.legend(loc="lower right", shadow=True, fontsize="x-large")

###############################################
# Show graph
# ----------

plt.show()

# %%
# .. image:: ../../_static/brake_maximum_temperature.png
#    :align: center
#    :alt: Brake Maximum Temperature

# %%
#    Brake Maximum Temperature

####################################################################################
# Close the session
# ==================================================================================

session.exit()

# sphinx_gallery_thumbnail_path = '_static/brake_surface_temperature-thumbnail.png'



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from ansys.fluent.core.examples import download_file


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=23.1")
def test_pure_meshing_mode(mixing_elbow_watertight_pure_meshing_session):
    pure_meshing_session = mixing_elbow_watertight_pure_meshing_session
    # check a few dir elements
    # n.b. 'field_data', 'field_info' need to
    # be eliminated from meshing sessions
    session_dir = dir(pure_meshing_session)
    for attr in ("fields", "meshing", "workflow"):
        assert attr in session_dir
    workflow = pure_meshing_session.workflow
    workflow_dir = dir(workflow)
    for attr in ("TaskObject", "InsertNewTask", "Workflow", "setState"):
        assert attr in workflow_dir
    import_geometry = workflow.TaskObject["Import Geometry"]
    import_geometry_dir = dir(import_geometry)
    for attr in ("AddChildToTask", "Arguments", "Execute", "setState"):
        assert attr in import_geometry_dir
    with pytest.raises(AttributeError):
        pure_meshing_session.switch_to_solver()


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=23.1")
def test_meshing_mode(new_meshing_session):
    meshing_session = new_meshing_session
    # check a few dir elements
    # n.b. 'field_data', 'field_info' need to
    # be eliminated from meshing sessions
    session_dir = dir(meshing_session)
    for attr in ("fields", "meshing", "workflow"):
        assert attr in session_dir
    assert meshing_session.workflow.InitializeWorkflow(
        WorkflowType="Watertight Geometry"
    )
    assert meshing_session.switch_to_solver()


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=23.1")
def test_meshing_and_solver_mode_exit(new_meshing_session):
    meshing_session = new_meshing_session
    solver_session = meshing_session.switch_to_solver()
    # Even if exit statement is invoked twice, only one is executed as the channel instance is shared
    meshing_session.exit()
    solver_session.exit()


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=23.1")
def test_meshing_mode_post_switching_to_solver(new_meshing_session):
    meshing_session = new_meshing_session
    meshing_session.switch_to_solver()
    # Post switching to solver session, meshing session specific attributes are unavailable
    with pytest.raises(AttributeError):
        meshing_session.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")


# def test_transfer_mesh_to_solvers(
#     new_pure_meshing_session, new_solver_session
# ):
#     mesh_file_name = download_file("mixing_elbow.msh.h5", "pyfluent/mixing_elbow")
#     pure_meshing_session = new_pure_meshing_session
#     pure_meshing_session.tui.file.read_mesh(mesh_file_name)
#     pure_meshing_session.tui.mesh.check_mesh()
#     mesh_info = pure_meshing_session.scheme_eval.string_eval(
#         "(%tg-length-of-entity-list)"
#     )
#     pure_meshing_session_cell_count = mesh_info.strip("( )").split()[3]
#
#     solver_session = new_solver_session
#     pure_meshing_session.transfer_mesh_to_solvers([solver_session], file_type="mesh")
#     solver_session.tui.mesh.check()
#     mesh_info = solver_session.scheme_eval.string_eval("(inquire-grids)")
#     solver_session_cell_count = mesh_info.strip("( )").split()[1]
#
#     assert pure_meshing_session_cell_count == solver_session_cell_count


def test_transfer_case_to_solvers(new_pure_meshing_session, new_solver_session):
    case_file_name = download_file("mixing_elbow.cas.h5", "pyfluent/mixing_elbow")
    pure_meshing_session = new_pure_meshing_session
    pure_meshing_session.tui.file.read_case(case_file_name)
    pure_meshing_session.tui.mesh.check_mesh()
    mesh_info = pure_meshing_session.scheme_eval.string_eval(
        "(%tg-length-of-entity-list)"
    )
    pure_meshing_session_cell_count = mesh_info.strip("( )").split()[3]

    solver_session = new_solver_session
    pure_meshing_session.transfer_mesh_to_solvers([solver_session], file_type="case")
    solver_session.tui.mesh.check()
    mesh_info = solver_session.scheme_eval.string_eval("(inquire-grids)")
    solver_session_cell_count = mesh_info.strip("( )").split()[1]

    assert pure_meshing_session_cell_count == solver_session_cell_count



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from ansys.fluent.core import examples
from ansys.fluent.core.filereader.case_file import SettingsFile as SettingsReader


def call_settings_reader(
    settings_file_name: str | None = None, expected: dict | None = None
):
    reader = SettingsReader(settings_file_name=settings_file_name)
    if expected is not None:
        assert reader.precision() == expected["precision"]
        assert reader.num_dimensions() == expected["num_dimensions"]
        assert reader.iter_count() == expected["iter_count"]
        assert {
            p.name: (p.numeric_value, p.units) for p in reader.input_parameters()
        } == expected["input_parameters"]
        assert {p.name: p.units for p in reader.output_parameters()} == expected[
            "output_parameters"
        ]


def call_settings_reader_static_mixer(
    settings_file_name: str | None = None,
):
    call_settings_reader(
        settings_file_name=settings_file_name,
        expected=dict(
            precision=2,
            num_dimensions=3,
            iter_count=100,
            input_parameters=dict(
                inlet1_temp=(300, "K"),
                inlet1_vel=(1, "m/s"),
                inlet2_temp=(350, "K"),
                inlet2_vel=(1, "m/s"),
            ),
            output_parameters={
                "outlet-temp-avg-op": "K",
                "outlet-vel-avg-op": "m s^-1",
            },
        ),
    )


def static_mixer_settings_file():
    return examples.download_file(
        "Static_Mixer_Params",
        "pyfluent/static_mixer",
        return_without_path=False,
    )


def test_settings_reader_static_mixer_h5():
    call_settings_reader_static_mixer(settings_file_name=static_mixer_settings_file())


def test_meshing_unavailable():
    reader = SettingsReader(settings_file_name=static_mixer_settings_file())
    with pytest.raises(AttributeError):
        reader.get_mesh()


def test_settings_reader_get_rp_and_config_vars():
    reader = SettingsReader(settings_file_name=static_mixer_settings_file())
    rp_vars = reader.rp_vars()
    assert rp_vars
    assert hasattr(rp_vars, "__getitem__")
    config_vars = reader.config_vars()
    assert config_vars
    assert hasattr(config_vars, "__getitem__")
    assert config_vars["rp-3d?"] is True
    assert reader.config_var("rp-3d?") is True
    assert reader.config_var.rp_3d__q() is True
    assert len(reader.rp_var.context.map_r17__plus()) == 53

    with pytest.raises(RuntimeError) as msg:
        reader.rp_var.defaults.pre_r19__dot0_early()

    with pytest.raises(ValueError) as msg:
        reader.config_var("rp-3d")
    assert (
        msg.value.args[0] == "'config-vars' has no attribute 'rp-3d'.\n"
        "The most similar names are: rp-3d?, rp-des?."
    )



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import ast

import pytest

from ansys.fluent.core import CODEGEN_OUTDIR


@pytest.mark.codegen_required
@pytest.mark.fluent_version("==25.1")
def test_settings_stub():
    # The type-stub files, which are generated for settings API, are parsed by the
    # intellisense engine while typing in editors like vscode. This test validates the
    # information contained in a type-stub file.
    version = "251"
    stub_file = CODEGEN_OUTDIR / "solver" / f"settings_{version}.pyi"
    assert stub_file.exists()
    with open(stub_file) as f:
        module_def = ast.parse(f.read())
    assert isinstance(module_def, ast.Module)
    assert any(isinstance(x, ast.ImportFrom) for x in module_def.body)
    class_def = next(
        x for x in module_def.body if isinstance(x, ast.ClassDef) and x.name == "export"
    )
    assert len(class_def.bases) > 0
    assigns = [x for x in class_def.body if isinstance(x, ast.AnnAssign)]
    for class_attr in [
        "_version",
        "fluent_name",
        "_python_name",
        "child_names",
        "command_names",
    ]:
        print(class_attr)
        assert any(x.target.id == class_attr for x in assigns)
    fn_def = next(x for x in class_def.body if isinstance(x, ast.FunctionDef))
    assert ast.get_docstring(fn_def)
    assert all(x.annotation for x in fn_def.args.args[1:])
